# (c) Roxy Corp. 2020-
# Roxy AI Inspect-Server API
from __future__ import annotations
import time
from struct import pack, unpack
from io import BytesIO

from .connection import Connection
from .com_definition import (
    SIGN_CODE,
    HEADER_SIZE,
    CommandStatus,
    CommandCode,
)

from logging import getLogger
log = getLogger(__name__)


class BaseCommand():
    # デフォルトの検査サーバの定義
    HOST = "127.0.0.1"
    """ デフォルトの接続先IPアドレス """
    PORT = 6945
    """ デフォルトの接続先ポート番号 """
    _REPLAY_NONE = False

    _CODE = None
    _connection = None
    _last_connection = Connection(HOST, PORT)
    data = b""
    _logging = True

    def __init__(self, connection: Connection = None, logging: bool = True):
        """ 基底コマンドクラス
        Args:
            connection:     通信対象のTCP接続
            logging:        送受信時のログ出力フラグ
        """
        self._connection = connection or self._connection or BaseCommand._last_connection
        if connection:
            # 接続設定をクラス変数に記録
            BaseCommand._last_connection = connection

        self.status = CommandStatus.NONE
        self.__send_time = None
        self.__recv_time = None
        self.__rest_size = None
        self._logging = logging

    def send(self, extension: bytes = b'', timeout: float = None):
        """ コマンドの送信

        Args:
            extension
                (bytes, option):    拡張送信データ
            timeout
                (None, default):    None はタイムアウト無し
                (float, option):    タイムアウト時間[sec]

        Raises:
            送信タイムアウト（socket.timeout）を含む通信例外を送出

        Note:
            extension はコマンドの最後に、メモリコピーを伴わずに送信する
            拡張データを指定できます。
        """
        data = self.data
        sock = self._connection.sock

        code = self._CODE
        size = len(data) + len(extension)
        status = CommandStatus.STS_REQUEST

        buffer = pack(f"< H L B B {len(data)}s", SIGN_CODE, size, code, status, data)

        if self._logging:
            log.debug(f"Send [{size:3,d} bytes]: {self}")

        self.__send_time = time.time()
        self.__recv_time = None
        sock.settimeout(timeout)
        sock.sendall(buffer)
        if extension:
            # 残りのタイムアウト時間を再計算してブロッキングモードの範囲で設定
            if timeout is not None:
                timeout = max(0.000001, timeout - (time.time() - self.__send_time))
            sock.settimeout(timeout)
            # 拡張データの送信
            sock.sendall(extension)

    def recv(self, recv_size: int = None, timeout: float = None):
        """ コマンドの受信

        Args:
            recv_size
                (None, default):    応答コマンド全体を受信
                (int, option):      指定されたデータBytesまで受信
            timeout
                (None, default):    None はタイムアウト無し
                (float, option):    タイムアウト時間[sec]
        Raises:
            送信タイムアウト（socket.timeout）を含む通信例外を送出

        Note:
            下記の属性を設定
            status (int):           返信コマンドステータス
            code (int):             返信コマンド番号
            rest_size (int):        フレームの未受信データサイズ
        """
        sock = self._connection.sock
        # タイムアウト設定
        pre_time = time.time()
        sock.settimeout(timeout)
        # ヘッダの読み込み
        buf = sock.recv(HEADER_SIZE)
        self.__recv_time = time.time()
        if len(buf) == 0:
            # サーバによる切断
            raise ConnectionResetError

        if len(buf) < HEADER_SIZE:
            # ゴミ受信のため破棄
            self.status = CommandStatus.ERR_INVALID_DATA_SIZE
            dat = ' '.join(f'{b:02x}' for b in buf)
            raise RuntimeError(f"{self} Receive invalid header size data: {len(buf)} bytes\nData: {dat}  {buf}")

        sign, size, code, status = unpack("< H L B B", buf[0:HEADER_SIZE])
        if sign != SIGN_CODE:
            # パケット種別チェック
            self.status = CommandStatus.ERR_INVALID_SIGNATURE
            dat = f'sign: {sign:04x}, size: {size:,d}, code: {code:02x}, status: {status:02x}'
            raise RuntimeError(f"{self} Receive invalid signe code: 0x{sign:04x}\n{dat}")

        # 受信ヘッダデータの検証
        if code and (code != self._CODE):
            self.status = CommandStatus.ERR_INVALID_COMMAND
            dat = f'sign: {sign:04x}, size: {size:,d}, code: {code:02x}, status: {status:02x}'
            raise RuntimeError(f"{self} Received command code:0x{code:02X} is mismatched 0x{self._CODE:02X}\n{dat}")

        try:
            self.status = status
        except ValueError:
            self.status = CommandStatus.ERR_UNKNOWN_EXCEPTION
            dat = f'sign: {sign:04x}, size: {size:,}, code: {code:02x}, status: {status:02x}'
            raise RuntimeError(f"{self} Receive invalid status: 0x{status:02x}\n{dat}")

        if recv_size and (recv_size < size):
            # フレームサイズが指定受信サイズ以上の場合は途中までを受信
            self.__rest_size = size - recv_size
            size = recv_size
            log.warning(f'{self} Received larger size packet: {size:,} > expectation: {recv_size:,} bytes')
        else:
            self.__rest_size = 0

        # コマンドデータの読み込み
        reply_data = b''
        if size > 0:
            # データ受信開始の時間を記録
            rcv_dat_start = time.time()
            with BytesIO(b'\0' * size) as buffer:
                while buffer.tell() < size:
                    # 残りのタイムアウト時間を再計算してブロッキングモードの範囲で設定
                    if timeout is not None:
                        now = time.time()
                        timeout = max(0.000001, timeout - (now - pre_time))
                        pre_time = now
                    sock.settimeout(timeout)

                    fragment = sock.recv(size - buffer.tell())

                    # データ受信開始の時間を記録
                    rcv_dat_time = (time.time() - rcv_dat_start) * 1000
                    received_size = buffer.tell() + len(fragment)
                    if rcv_dat_time >= 500:
                        # 500ms 以上掛かっている場合は警告ログを出力
                        log.warning(
                            f'{self}\n Long receiving time: {rcv_dat_time:,} ms ({received_size:,} / {size:,} bytes) '
                        )
                    if received_size > size:
                        # バッファで最初に確保したサイズ以上のデータを受信した場合
                        log.warning(
                            f'{self}\n Received larger size data: {received_size:,} / {size:,} bytes'
                        )

                    if len(fragment) == 0:
                        # サーバによる切断
                        raise ConnectionResetError
                    buffer.write(fragment)
                reply_data = buffer.getvalue()

            # 受信コマンドデータのデコード
            try:
                self._decode_reply(reply_data)
            except Exception:
                self.status = CommandStatus.ERR_FIALED_PARSE_DATA
                log.exception('received invalid data')

        if self._logging:
            log.debug(f"Recv [{size:3,d} bytes]: {self}")

        if self._status.is_error_reply:
            reply_data = b''

    def _decode_reply(self, reply: bytes):
        if reply:
            # データが存在する場合は各コマンドでオーバライドする
            raise NotImplementedError

    def run(self, timeout: float = None):
        """ コマンドの送受信
        Args:
            timeout
                (None, default):    None はタイムアウト無し
                (float, option):    タイムアウト時間[sec]
        Raises:
            送信タイムアウト（socket.timeout）を含む通信例外を送出
        """
        pre_time = time.time()
        # コマンドの送信＆受信
        self.send(timeout=timeout)

        if self._REPLAY_NONE:
            # 応答不要コマンドは受信完了を設定する
            self.__recv_time = time.time()
            self.status = CommandStatus.STS_REPLY_ACK
            return b''

        if timeout is not None:
            # 残りのタイムアウト時間を再計算してブロッキングモードの範囲で設定
            now = time.time()
            timeout = max(0.000001, timeout - (now - pre_time))
            pre_time = now
        self.recv(timeout=timeout)

    @property
    def is_received(self) -> bool:
        """ 応答を受信しているかのフラグ
        """
        return (self.__recv_time is not None)

    @property
    def is_received_ack(self) -> bool:
        """ 正常応答を受信しているかのフラグ
        """
        return (self._status and self._status.is_ack)

    @property
    def process_time(self) -> float:
        """ コマンドの要求～応答の処理時間 [ms]
        """
        proc_time = None
        if self.is_received:
            proc_time = (self.__recv_time - self.__send_time) * 1000
            proc_time = round(proc_time, 3)
        return proc_time

    def __str__(self) -> str:
        string = f"{str(self.code)} "
        if not self.is_received:
            string += "<REQUEST>"
        else:
            string += "<REPLY"
            if self._status.is_error_reply:
                string += f"-{str(self.status)}"
            string += ">"
            if self.is_received:
                string += f" ({self.process_time} ms)"
        return string

    # 読み出し専用変数のアクセス定義

    @property
    def code(self) -> CommandCode:
        """ コマンドのコード番号
        """
        return CommandCode(self._CODE)

    @property
    def last_connection(self) -> Connection:
        """ 前回接続に利用した Connection
        """
        return self._last_connection

    @property
    def connection(self) -> Connection:
        """ 接続中の Connection
        """
        return self._connection

    @property
    def status(self) -> CommandStatus:
        """ コマンドの送信／返信状態
        """
        return CommandStatus(self._status)

    @status.setter
    def status(self, val: int):
        """ コマンドの送信／返信状態の設定
        """
        self._status = CommandStatus(val)

    @property
    def rest_size(self) -> int:
        """ 応答コマンドの未受信データの残りサイズ
        """
        self.__rest_size



# (c) Roxy Corp. 2020-
# Roxy AI Inspect-Server API
from __future__ import annotations
from typing import List, Union
import struct
from datetime import datetime

from .connection import Connection
from .com_definition import (
    Probability,
    LabeledProbability,
    CommandCode,
    Judgment,
    ProbabilityType,
)
from .com_base import (
    BaseCommand,
    HEADER_SIZE,
)
from .inspect_image import ImageFormat, InspectImage


class InspectCommand(BaseCommand):

    _CODE = CommandCode.INSPECT
    PROB_SIZE = 13
    LABELED_PROB_SIZE = 14
    PROB_OFFSET = 20 - HEADER_SIZE

    # Requestパラメータ
    _inspect_id = None
    _model_id = None
    _data_format = None
    _image_data = b''
    # Replayパラメータ
    _result = Judgment.FAILED
    _prob_size = 0
    _prob_list: list[Probability] = []

    # ログの詳細出力
    verbose = False

    def __init__(
        self,
        inspect_id: int,
        model_id: int,
        data_format: ImageFormat = None,
        image_data: Union[bytes, InspectImage] = b'',
        connection: Connection = None,
        logging: bool = True,
    ):
        """ Inspect コマンド制御

        Args:
            inspect_id:     検査番号
            model_id:       モデル番号
            data_format:    送信する検査画像フォーマット
                None:           image_data のフォーマット利用
                ImageFormat:    送信フォーマットを指定
            image_data:     検査画像データ
            connection:     通信対象のTCP接続
            logging:        送受信時のログ出力フラグ

        Raises:
            TypeError:      data_format が None で image_data が bytes
            ValueError:     data_format が NDARRAY
        """
        super().__init__(connection=connection, logging=logging)
        # データの格納フォーマットチェック
        if (data_format is None) and not issubclass(type(image_data), InspectImage):
            raise TypeError('image data_format is not specified')
        # 要求データの設定
        if inspect_id is None:
            self._inspect_id = self.get_datetime_id()
        else:
            self._inspect_id = inspect_id
        self._model_id = model_id
        self._data_format = data_format or image_data.format
        if self._data_format == ImageFormat.NDARRAY:
            raise ValueError('ImageFromat NDARRAY is not supported')

        # データの送付方法設定
        self._image_data = image_data
        if (
            (self._data_format == ImageFormat.RAW)
            and issubclass(type(image_data), InspectImage)
            and image_data.mmap_path
        ):
            # メモリマップドファイルとしてパスを渡す
            mpath = str(image_data.mmap_path).encode('utf-8')
            data = struct.pack(
                f'< B {len(mpath)}s', 0xFF, mpath
            )
            self._extension = b''
        else:
            # データを拡張データとして渡す
            # self._image_data = image_data.get_ndarray()
            self._extension = self._image_data
            data = b''
        self.encode_data(data)

        # 応答の変数定義はクラス定義を引き継ぎ
        # self._result = Judgment.FAILED
        # self._prob_size = 0
        # self._prob_list = []
        self._prob_type = None

    def encode_data(self, data: bytes = None):
        """ Inspect コマンドの送信データ生成
        """
        send_params = (
            self._inspect_id,
            self._model_id,
            self._data_format,
        )
        if data:
            self.data = struct.pack(
                f'< Q B B {len(data)}s', *send_params, data
            )
        else:
            self.data = struct.pack('< Q B B', *send_params)

    def send(self, timeout: float = None):
        """ Inspect コマンドの要求を送信

        Args:
            timeout:
                (None, default):    None はタイムアウト無し
                (float, option):    タイムアウト時間[sec]

        Raises:
            送信タイムアウト（socket.timeout）を含む通信例外を送出
        """
        super().send(extension=self.extension, timeout=timeout)

    def recv(self, timeout: float = None):
        """ Inspect コマンドの応答を受信

        Args:
            timeout:
                (None, default):    None はタイムアウト無し
                (float, option):    タイムアウト時間[sec]

        Raises:
            送信タイムアウト（socket.timeout）を含む通信例外を送出
            RuntimeError    不正な応答受信
        """
        super().recv(timeout=timeout)

    def _decode_reply(self, reply: bytes):
        """ Inspect コマンドの応答内容確認

        Args:
            reply (bytes):      受信応答データ（ヘッダ以外）
        """
        if not reply:
            self._result = Judgment.FAILED
            self._prob_size = 0
            self._prob_list = []
            return

        inspect_id, result, prob_type, prob_size = struct.unpack(
            '< Q B B H',
            reply[:self.PROB_OFFSET]
        )
        prob_data = reply[self.PROB_OFFSET:]

        # 受信データのチェック
        if self._inspect_id != inspect_id:
            raise RuntimeError(
                f'mismatched inspect_id '
                f'recv:{inspect_id}(=0x{inspect_id:#016X}) != '
                f'send:{self._inspect_id}(=0x{self._inspect_id:#016X})')

        # 受信データの格納
        self._result = Judgment(result)
        self._prob_type = prob_type
        self._prob_size = prob_size
        self._prob_list = []
        if prob_type == ProbabilityType.LABELED:
            for offset in range(
                0, prob_size * self.LABELED_PROB_SIZE, self.LABELED_PROB_SIZE
            ):
                x1, y1, x2, y2, typ, label, prob = struct.unpack(
                    '< H H H H B B f',
                    prob_data[offset:offset + self.LABELED_PROB_SIZE]
                )
                prob = LabeledProbability(x1, y1, x2, y2, typ, label, prob)
                self._prob_list .append(prob)
        else:
            for offset in range(0, prob_size * self.PROB_SIZE, self.PROB_SIZE):
                x1, y1, x2, y2, typ, prob = struct.unpack(
                    '< H H H H B f',
                    prob_data[offset:offset + self.PROB_SIZE]
                )
                prob = Probability(x1, y1, x2, y2, typ, prob)
                self._prob_list .append(prob)

    @staticmethod
    def get_datetime_id() -> int:
        """ デフォルトで利用する日時算出の検査ID取得

        Returns:
            int:        日時から算出した検査ID
        """
        dt = datetime.now()
        dtid = int(dt.strftime('%Y%m%d%H%M%S%f')) // 1000
        return dtid

    def __str__(self) -> str:
        if self._extension:
            image_size = self._extension.get_size(self._data_format)
        else:
            image_size = self._image_data.get_size(self._data_format)
        string = (
            f'{super().__str__()} '
            f'InspectID: {self._inspect_id}(=0x{self._inspect_id:016X}), '
            f'ModelID: {self._model_id}, '
            f'DataFormat: 0x{self._data_format:02X}, '
            f'ImageData: {image_size:,d} bytes -> '
        )
        if self.is_received_ack:
            string += (
                f'Result: {str(self._result)}, '
                f'ProbabilityType: {self._prob_type}, '
                f'ProbabilitySize: {self._prob_size}, '
                f'ProbabilityList: {self._prob_size} items '
            )
            if self.verbose:
                for prob in self._prob_list:
                    string += '\n    ' + str(prob)
        return string

    # 読み取り専用変数の定義
    @property
    def inspect_id(self) -> int:
        return self._inspect_id

    @property
    def model_id(self) -> int:
        return self._model_id

    @property
    def data_format(self) -> ImageFormat:
        return self._data_format

    @property
    def image_data(self) -> InspectImage:
        return self._image_data

    @property
    def extension(self) -> bytes:
        if self._extension:
            return self._extension.get_image(self._data_format)
        return self._extension

    @property
    def result(self) -> Judgment:
        return self._result

    @property
    def prob_size(self) -> int:
        return self._prob_size

    @property
    def prob_list(self) -> List[Probability]:
        return self._prob_list



# (c) Roxy Corp. 2020-
# Roxy AI Monitor API
from __future__ import annotations
from pathlib import Path
import json5 as json
from logging import getLogger
from dataclasses import dataclass

log = getLogger(__name__)


# 共通のラベル情報定義
OK_LABEL_NAME = 'OK'
OK_LABEL_COLOR = 'R=256,G=256,B=256'
OTHER_LABEL_NAME = 'Other'
OTHER_LABEL_COLOR = 'R=127,G=127,B=127'


def _color2bgr(color: str) -> tuple[int, int, int]:
    """ カラー定義文字列からRGB値への変換
    Args:
        color:  RGB値定義文字列
            例) R=256,G=256,B=256
    Return:
        tuple[R, G, B]      変換値
        None:               変換失敗
    """
    try:
        rs, gs, bs = color.split(',')
        ri = int(rs.replace('R=', ''))
        gi = int(gs.replace('G=', ''))
        bi = int(bs.replace('B=', ''))
        ret = (bi, gi, ri)
    except Exception:
        ret = None
    return ret


@dataclass
class ModelLabelInfo():
    """ モデル毎のラベル情報
    """
    id: int
    name: str
    color: str

    def __post_init__(self):
        self.bgr = _color2bgr(self.color)

    def __str__(self) -> str:
        text = f'label ({self.id}): "{self.name}" {self.color}'
        return text

    @classmethod
    def OK(cls) -> ModelLabelInfo:
        return ModelLabelInfo(0, OK_LABEL_NAME, OK_LABEL_COLOR)


class ProductLabelInfo():
    """ 製品全体の統合ラベル情報
    """
    # ラベル情報（0番は利用しない）
    _label_list = []
    LABEL_ID_OK = 0

    # 違和感検知（その他の不良）用のラベル
    _other_label = None
    _ok_label = None

    @classmethod
    def clear(cls):
        cls._label_list = []
        cls._other_label = None
        # 正常ラベルはデフォルトで登録しておく
        cls._ok_label = ProductLabelInfo(
            cls.LABEL_ID_OK, OK_LABEL_NAME, OK_LABEL_COLOR
        )
        cls._label_list.append(cls._ok_label)

    @classmethod
    def set_info(
        cls, model_name: str, index: int, label_name: str, color: str
    ) -> int:
        """ ラベルを設定してIDを取得
        Args:
            model_name:     モデル名
            index:          モデルにおけるラベルのインデックス番号
            label_name:     ラベル名
            color:          色情報
        Returns:
            ラベルの統一番号
        """
        info = ProductLabelInfo(index, label_name, color)
        if info in cls._label_list:
            # 既に同じラベルがあればそれに差替え
            i = cls._label_list.index(info)
            info = cls._label_list[i]
            label_id = info._label_id
        else:
            # 新規なら追加登録
            label_id = len(cls._label_list)
            info._label_id = label_id
            cls._label_list.append(info)
        # モデル名を登録
        info._model_name.append(model_name)

        if cls._other_label:
            # OTHERラベルがある場合は最後のラベルとして設定
            cls._other_label._label_id = len(cls._label_list)

        return label_id

    @classmethod
    def set_other(cls, model_name: str, index: int):
        """ 違和感検知用に「その他の不良」のラベルを設定する
        Args:
            model_name:     モデル名
            index:          モデルにおけるOTHERラベルのインデックス番号
        Note:
            その他の不良ラベルは常にラベルリストの最後に設定
        """
        if not cls._other_label:
            # ラベルが未登録ならば新規作成
            cls._other_label = ProductLabelInfo(
                index, OTHER_LABEL_NAME, OTHER_LABEL_COLOR
            )
        # モデル名を登録
        cls._other_label._model_name.append(model_name)
        # 最後のラベルとしてIDを設定
        cls._other_label._label_id = len(cls._label_list)

    @classmethod
    def get(cls, label_id: int):
        """ ラベルIDから情報を取得
        Args:
            label_id:   ProductLabelInfoのラベルの統一番号
        """
        if (label_id < 0) or (len(cls._label_list) <= label_id):
            ret = cls._other_label
        else:
            ret = cls._label_list[label_id]
        return ret

    @classmethod
    def get_list(cls, include_ok: bool = False) -> list:
        """ その他の不良を含めた全てのリストを返す
        """
        if include_ok:
            # 正常ラベルを含める場合
            all_list = cls._label_list.copy()
        else:
            # 正常ラベルを含めない場合
            all_list = cls._label_list[1:].copy()

        if cls._other_label:
            # 違和感検知用に「その他の不良」ラベルがあれば追加
            all_list.append(cls._other_label)

        return all_list

    @classmethod
    def get_other_id(self) -> int:
        """ 違和感検知用のOTHERラベルの統一番号を取得
        Note:
            常に全モデルのラベルの最後に追加される
        """
        return len(self._label_list)

    def __init__(self, index: int, name: str, color: str):
        """ ラベル情報
        Args:
            index:             ラベル番号（モデル内の番号）
            name:           ラベル名
            color:          色情報
        """
        # 統合ラベル番号
        self._index = index
        self._name = name
        self._color = color

        # 統合ラベル番号
        self._label_id = 0
        # このラベルを含むモデル名
        self._model_name = []

    def __eq__(self, o: ProductLabelInfo) -> bool:
        return (
            (self._index == o._index)
            and (self._name == o._name)
            and (self._color == o._color)
        )

    @property
    def id(self) -> int:
        return self._label_id

    @property
    def model_name(self) -> int:
        return self._model_name

    @property
    def index(self) -> int:
        return self._index

    @property
    def name(self) -> str:
        return self._name

    @property
    def color(self) -> str:
        return self._color

    @property
    def bgr(self) -> tuple[int, int, int]:
        return _color2bgr(self._color)

    def __str__(self) -> str:
        text = f'label ({self._label_id}): "{self._model_name}" '
        text += f'{self._index:2d}) {self._name} {self._color}'
        return text


class LabelManager():
    """ ラベルとモデルの関連を管理する
    """
    _LABELS_FILE = 'labels.txt'
    _MODEL_CONFIG_FILE = 'config.json'
    _INSPECT_CONFIG_FILE = 'inspect_config.json'

    @classmethod
    def get(cls, product_folder: str, model_list: list[str] = None) -> LabelManager:
        """ 製品全体のラベル情報を取得する
        Args:
            product_folder:     製品情報フォルダ
            model_list:         使用モデル名のリスト
        Return:
            LabelManager
        """
        try:
            ProductLabelInfo.clear()
            label_info = cls(product_folder, model_list)
        except Exception as e:
            log.exception(f'{cls.__name__}: failed to load label info\n{e}')
            label_info = None
        return label_info

    def __init__(self, product_folder: str, model_list: list[str] = None):
        """ モデル毎のラベル情報を管理するクラス
        Args:
            product_folder (str or Path):   製品フォルダ
            model_list:         使用モデル名のリスト
        Note:
            製品名フォルダ配下のinspect_config.jsonからモデル名を取得し、
        """
        self._product_path = Path(product_folder)

        self._label_set = set()
        self._model_list = model_list
        # モデル毎の統合ラベル番号辞書
        self._product_label_dict: dict[str, dict[int, ProductLabelInfo]] = {}
        # モデル毎のラベル番号辞書
        self._model_label_dict: dict[str, ModelLabelInfo] = {}

        if not self._product_path.exists():
            # ファイルが存在しない
            log.error(f'Not found inspect_config.json. {self._product_path}')
            raise RuntimeError
        self._load_all_model_label()

    def _load_all_model_label(self):
        """ 製品フォルダ配下の全てのモデルのラベル情報を読込
        """
        model_path_list = list(
            path for path in self._product_path.iterdir()
            if path.is_dir()
        )
        labels_path_list = list(
            md / self._LABELS_FILE for md in model_path_list
        )

        # 違和感検知用のOTHERラベルのインデックス（全てのモデルで共通化のため別管理）
        other_label = {}

        for label_path in labels_path_list:
            if not label_path.exists():
                # ラベルファイルが無ければ無効フォルダとしてスキップ
                continue
            model_name = label_path.parent.name
            if self._model_list and (model_name not in self._model_list):
                # 未使用のモデルならスキップ
                continue

            # モデルラベル情報の生成または取得
            model_dict = self._model_label_dict.get(
                model_name, {0: ModelLabelInfo.OK()}
            )
            # モデル毎のラベル情報の読み出し
            try:
                data = label_path.read_text(encoding='utf_8_sig', errors='ignore')
                labels = {}
                # OKがラベルID=0となるため不良ラベルは1開始
                index = 1
                for line in data.rstrip('\n').split('\n'):
                    # ラベル定義の行から値を取得
                    parts = line.rstrip('\n').split('\t')
                    # ラベル定義のindexは無視
                    # index = int(parts[0])
                    name = parts[1]
                    color = parts[2]
                    used = parts[3]

                    if used == 'unused':
                        # 未使用ラベルはindexの対象外
                        continue

                    # 読み込んだ情報を登録
                    label_info = ProductLabelInfo.set_info(
                        model_name, index, name, color
                    )
                    labels[index] = label_info
                    model_dict[index] = ModelLabelInfo(index, name, color)
                    index += 1

            except Exception as e:
                log.warning(f'Failed to load: {label_path}, {e}')
                # 失敗したファイルは無視して処理継続

            # 違和感検知の有効／無効読み出し
            cfg_path = label_path.parent / self._MODEL_CONFIG_FILE
            try:
                if cfg_path.exists():
                    # モデル設定から Anomal の有効／無効を取得
                    data = cfg_path.read_text(encoding='utf_8_sig ', errors='ignore')
                    dic: dict = json.loads(data)
                    if dic.get('anormal_param', {}).get('enable_anormal', False):
                        # 違和感検知が有効ならばOTHERラベルを登録
                        ProductLabelInfo.set_other(model_name, index)
                        other_label[model_name] = index
                        model_dict[index] = ModelLabelInfo(
                            index, OTHER_LABEL_NAME, OTHER_LABEL_COLOR
                        )
            except Exception as e:
                log.warning(f'Failed to load model config: {cfg_path}, {e}')

            # モデル毎の辞書に登録
            self._product_label_dict[model_name] = labels
            self._model_label_dict[model_name] = model_dict
            log.debug(f'Loaded {len(labels)} labels for model "{model_name}".')

        # 違和感検知用のラベル番号を各モデルに登録
        for model_name, index in other_label.items():
            self._product_label_dict[model_name][index] = ProductLabelInfo.get_other_id()

        log.debug(f'loaded total {len(ProductLabelInfo.get_list())} labels.')

    def get_info_json(self) -> str:
        """ 統合ラベルの情報取得
        Returns:
            全モデルの統合ラベル情報を表すJSONデータ
        """
        # モニター設定情報の取得
        conf_path = (self._product_path / self._INSPECT_CONFIG_FILE).resolve().absolute()
        try:
            data = conf_path.read_text(encoding='utf-8', errors='ignore')
            dic: dict = json.loads(data)
            monitor_conf = dic.get('monitor', {})
        except Exception as e:
            log.warning(f'failed to load monitor config: {conf_path}\n{repr(e)}')

        # モニター情報の読込
        title_list = []
        screen_list = monitor_conf.get('screen', [])
        for screen in screen_list:
            sc_id = screen.get('screen_id')
            sc_title = screen.get('title')
            if sc_title:
                title_list.append(sc_title)
            else:
                log.debug(f'{self}: screen {sc_id} does not have "title"')

        # 全ラベルの情報リスト
        label_list = []
        for label in ProductLabelInfo.get_list():
            label_list.append({
                'label_id': label.id,
                'label_name': label.name,
                'color': label.color,
            })

        all_label_info = {
            'label': label_list
        }

        label_json = json.dumps(all_label_info, ensure_ascii=False)
        return label_json

    def get_id(self, model_name: str, label_index: int) -> int:
        """ モデル名とモデル定義のラベル番号から統合のラベル番号取得
        Args:
            model_name:     モデル名
            label_index:    モデルデータ内に定義されたラベル番号
        Returns:
            全モデルで統一化されたラベル番号
        """
        labels = self._product_label_dict.get(model_name, {})
        label_id = labels.get(label_index)
        return label_id

    def get_model_label(self, model_name: str) -> dict[int, ModelLabelInfo]:
        """ モデル名からモデル毎のラベル情報辞書を取得する
        Args:
            model_name:     モデル名
        Returns:
            モデル毎のラベル情報辞書
        """
        return self._model_label_dict.get(model_name, {})

    @property
    def label_list(self) -> list[ProductLabelInfo]:
        return ProductLabelInfo.get_list()



# (c) Roxy Corp. 2020-
# Roxy AI Inspect-Server API
from __future__ import annotations
from typing import Union
import json

from .connection import Connection
from .com_definition import CommandCode
from .com_base import BaseCommand


class InitializeCommand(BaseCommand):

    _CODE = CommandCode.INITIALIZE
    send_json = ''

    def __init__(
        self,
        send_json: str = None,
        product: str = None,
        model_list: Union[list[str], tuple[str, str]] = [],
        connection: Connection = None,
        logging: bool = True,
    ):
        """ Initialize コマンド制御

        Args:
            send_json:      初期設定をJSONデータで定義
            product:        初期設定のproduct名（設定パス）
            model_list:     初期設定のモデル名リスト
            connection:     通信対象のTCP接続
            logging:        送受信時のログ出力フラグ

        Note:
            JSONデータか、product, model_list指定の何れかを選択する。

        Example:
            code-block:: markdown
            ```python
            from roxyai-api import InitializeCommand

            com = InitializeCommand(product='ProductName', model_list=['ModelName'])
            com.run()
            if com.status.is_error_reply:
                raise RuntimeError(str(com))
            ```
        """
        super().__init__(connection=connection, logging=logging)
        # 要求データの設定
        if send_json:
            self.product = None
            self.model_list = []
            self.send_json = send_json
        else:
            self.product = product
            self.model_list = model_list
            self.__encode_json()
        self.data = self.send_json.encode('utf-8')

    def append_model(self, model_name: str, group_name: str = None):
        """ モデルを追加する（オプションで画像グループ指定を行う）
        """
        if self.product is None:
            # jsonコードの未解釈の場合は一度解釈する
            dic = json.loads(self.send_json, encoding='utf-8')
            self.product = dic['Product']
            self.model_list = dic['ModelList']

        if group_name:
            self.model_list.append((str(model_name), str(group_name)))
        else:
            self.model_list.append(str(model_name))
        self.__encode_json()

    def __encode_json(self) -> str:
        model_list = []
        for model in self.model_list:
            if type(model) in (list, tuple):
                if len(model) >= 2:
                    model = [str(model[0]), str(model[1])]
                else:
                    model = str(model[0])
            else:
                model = str(model)
            model_list.append(model)
        dic = {
            'Product': str(self.product),
            'ModelList': model_list
        }
        self.send_json = json.dumps(dic, ensure_ascii=False)

    def __str__(self):
        string = (
            f'{super().__str__()} '
            f'SendJson: {self.send_json} '
            f'{len(self.send_json)} bytes ->'
        )
        return string



# (c) Roxy Corp. 2020-
# Roxy AI Coordinator
from __future__ import annotations
from typing import Any
from struct import pack, unpack
from time import perf_counter_ns
from io import BytesIO
from socket import socket
from logging import getLogger

from .com_definition import (
    SIGN_CODE,
    HEADER_SIZE,
    CommandStatus,
)
from .svr_base_handler import ServerBaseHandler

log = getLogger(__name__)


class ServerCommandManager:

    def __init__(
        self,
        sock: socket,
        handlers: dict[int, ServerBaseHandler],
        context: Any = None,
    ):
        """ 逐次コマンド処理クラス
        Args:
            connection:     接続管理クラス
            handlers:       ハンドラ辞書
            context:        ハンドラに渡すオプション情報
        """
        self._socket = sock
        self.handlers = handlers
        self.context = context

        self._cur_code = None
        self._cur_handler = None
        self._last_handler = None
        self._recv_data = None
        self._repeat_count = 0

        self.__recv_time = None
        self.__send_time = None

    def start(self):
        """ 処理開始
        Note:
            この関数は複数コマンドの並列処理には対応していない
        """
        while True:
            self._recv()
            self._dispatch()

    def _recv(self):
        """ コマンドの受信
        Note:
            下記のクラス属性を設定
            code (int)      受信要求コマンド番号
        """
        # 受信開始時刻の記録
        self.__recv_time = perf_counter_ns()
        self.__send_time = None

        # ヘッダの読み込み
        buf = self._socket.recv(HEADER_SIZE)
        if len(buf) == 0:
            # クライアントによる切断
            raise ConnectionResetError

        if len(buf) < HEADER_SIZE:
            # ゴミ受信のため破棄
            log.warning(f"Receive invalid header size data: {len(buf)} bytes")

        sign, size, code, status = unpack("< H L B B", buf[0:HEADER_SIZE])
        if sign != SIGN_CODE:
            # パケット種別チェック
            raise RuntimeWarning(f"Receive invalid signe code: 0x{sign:04x}")

        if status != CommandStatus.STS_REQUEST:
            # 不正なステータス受信
            raise RuntimeWarning(f"Receive invalid status: 0x{status:02x}")

        self._cur_code = code

        # コマンドデータの読み込み
        with BytesIO(b"\0" * size) as buffer:
            while buffer.tell() < size:
                fragment = self._socket.recv(size - buffer.tell())
                if len(fragment) == 0:
                    # クライアントによる切断
                    raise ConnectionResetError
                buffer.write(fragment)
            self._recv_data = buffer.getvalue()

        return

    def _dispatch(self):
        """ コマンドによってリクエスト処理を振り分ける
        """
        status = None
        try:
            code = self._cur_code
            hdl_cls = self.handlers.get(code)
            if hdl_cls:
                # コマンド処理ハンドラのインスタンス生成
                self._cur_handler: ServerBaseHandler = hdl_cls(
                    self._reply, self.context
                )
                # 各コマンドハンドラのデコード呼び出し
                self._cur_handler.decode(self._recv_data)
                log.debug(f">> Dispatch command {self._cur_handler}")
                # ハンドラ処理の実行
                self._cur_handler.proceed()
            else:
                log.warning(f">> Unkonw command ({code:02X}x)")
                status = CommandStatus.ERR_INVALID_COMMAND

        except Exception:
            # コマンド処理で未処理の例外発生の場合はエラー返信
            status = CommandStatus.ERR_UNKNOWN_EXCEPTION
            log.exception(f"Faild command {code:02X}x")

        if status is not None:
            # エラーで即時返信の場合
            self._reply(CommandStatus(status))

    def _reply(self, status: CommandStatus, data: bytes = b"", extension: bytes = b""):
        """ 応答コマンドの送信
        Args:
            status (CommandStatus): 応答ステータス
            data (bytes):           応答コマンドデータ
            extension (bytes):      応答コマンド追加送信データ
        """
        code = self._cur_code
        handler = self._cur_handler

        # 返信ステータスの確認
        if status.is_error_reply:
            # エラー応答はデータ無し
            data = b""
            extension = b""
        size = len(data) + len(extension)

        # 応答コマンドのデータ生成
        buffer = pack(f"< H L B B {len(data)}s", SIGN_CODE, size, code, status, data)

        self._socket.sendall(buffer)
        if extension:
            self._socket.sendall(extension)

        # 送信完了時間の記録
        self.__send_time = perf_counter_ns()

        # コマンド完了のログ出力
        last_handler = self._last_handler
        if (code, status) == last_handler:
            log.debug(f"<< Complete command {handler}")
            self._repeat_count += 1
        else:
            if self._repeat_count > 0:
                # 連続で同じコマンドは出力しない
                log.info(
                    f"<< Complete command {last_handler} " f"x {self._repeat_count} times"
                )
            log.info(f"<< Complete command {handler}")
            self._last_handler = handler
            self._repeat_count = 0

    @property
    def process_time(self) -> float:
        """ コマンドの要求～応答の処理時間 [ms]
        """
        time_ms = None
        if self.__send_time and self.__recv_time:
            time_ms = (self.__send_time - self.__recv_time) / 1000000
        return time_ms



# (c) Roxy Corp. 2020-
# Roxy AI Inspect-Server API
from __future__ import annotations

from .connection import Connection
from .com_definition import CommandCode
from .com_base import BaseCommand


class TerminateCommand(BaseCommand):
    _CODE = CommandCode.TERMINATE

    def __init__(self, connection: Connection = None, logging: bool = True):
        """ Terminate コマンド制御
        Args:
            connection:     通信対象のTCP接続
            logging:        送受信時のログ出力フラグ
        """
        super().__init__(connection=connection, logging=logging)
        # 要求データの設定
        self.data = b''



# (c) Roxy Corp. 2020-
# Roxy AI Analyze-Server API
from __future__ import annotations
import struct

from .connection import Connection
from .com_definition import CommandCode
from .com_base import BaseCommand, HEADER_SIZE


class GetImageCommand(BaseCommand):

    _CODE = CommandCode.GET_IMAGE
    PROB_OFFSET = 14 - HEADER_SIZE

    def __init__(
        self,
        inspect_id: int,
        connection: Connection = None,
        logging: bool = True,
    ):
        """ GetImage コマンド制御
        Args:
            inspect_id:     取得対象の検査番号
            connection:     通信対象のTCP接続
            logging:        送受信時のログ出力フラグ
        """
        super().__init__(connection=connection, logging=logging)
        self.inspect_id = inspect_id
        self.data = struct.pack('< Q', inspect_id)

        self.data_format = None
        self.image_data = None

    def _decode_reply(self, reply: bytes):
        """ Inspect コマンドの応答内容確認

        Args:
            reply (bytes):      受信応答データ（ヘッダ以外）
        """
        self.data_format, = struct.unpack('< B', reply)
        image_data = b''
        size = self.rest_size
        if size > 0:
            # 画像データの受信
            while len(image_data) < size:
                image_data += self.connection.sock.recv(size - len(image_data))
        self.image_data = image_data

    def __str__(self) -> str:
        string = (
            f'{super().__str__()} '
            f'InspectID: {self.inspect_id}(=0x{self.inspect_id:016X}) -> '
        )
        if self.is_received_ack:
            string += (
                f'DataFromat: 0x{self.data_format:02X}, '
                f'ImageData: {len(self.image_data):,d} bytes '
                f'({self.process_time} ms)'
            )
        return string



# (c) Roxy Corp. 2020-
# Roxy AI Inspect-Server API
from __future__ import annotations
from threading import current_thread
from socket import socket, AF_INET, SOCK_STREAM, timeout, errno
from socket import error as socket_error
from threading import Lock
from time import sleep

import logging.config

log = logging.getLogger(__name__)


class Connection():

    # クラス変数
    # 接続中のインスタンスを管理
    _dict = {}
    _host = None
    _port = None

    # 検査サーバのデフォルト定義
    HOST = "127.0.0.1"
    PORT = 6945

    # 接続管理定数
    RETRY_INTERVAL = 0.2  # sec

    def __new__(cls, *args, **kwargs):
        # 同じスレッドの同じサーバへの接続はインスタンスを使いまわす
        host = None
        port = None
        try:
            host = args[0]
            port = args[1]
        except IndexError:
            host = kwargs.get('host')
            port = kwargs.get('port')
        host = host or cls.HOST
        port = port or cls.PORT
        thread = current_thread()
        instance = cls._dict.get((host, port, thread))
        if not instance:
            instance = super().__new__(cls)
            cls._dict[(host, port, thread)] = instance
        return instance

    def __init__(
        self,
        host: str | None = None,
        port: int | None = None,
        retry: int | None = None,
        echo_check: bool = True,
    ):
        """ Inspect-Serverへの接続管理を行うクラス

        Args:
            host:
                None:    クラスのデフォルト接続先を利用
                str:     接続先ホストIPv4アドレス
            port:
                None:    クラスのデフォルト接続ポートを利用
                int:     接続先ホストTCPポート
            retry:                  サーバがビジーの場合に接続をリトライする回数
                None:    リトライ回数制限無し
                int:     指定リトライ回数
            echo_check:  接続時のエコーコマンドによるチェックの要否
        """
        if self._host is None:
            # 初回のみ初期化する
            self._host = host or self.HOST
            self._port = port or self.PORT
            self._lock = Lock()
            self._sock = None
        self.retry = retry
        self.echo_check = echo_check

    def __enter__(self) -> Connection:
        """ 接続開始 for with構文

        Returns:
            Connection:     接続した Connection のインスタンス
        """
        self.open(retry=self.retry)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """ 接続完了 for with構文
        """
        self.close()

    def open(self, retry: int | None = None):
        """ 接続開始

        Args:
            retry:                  サーバがビジーの場合に接続をリトライする回数
                (None, default):    リトライ回数制限無し
                (int, option):      指定リトライ回数
        """
        if not self._sock:
            # log.debug(f"wait lock {id(self)}")
            # self._lock.acquire()
            # log.debug(f"acuire lock {id(self)}")

            count = 0
            self._sock = socket(AF_INET, SOCK_STREAM)
            while True:
                try:
                    self._sock.connect((self._host, self._port))
                    if self.echo_check:
                        # Echoコマンドでの接続確認
                        # モジュールの依存関係のため実行時に読み込み
                        from .com_echo import EchoCommand
                        com = EchoCommand(b"", connection=self)
                        com.send(timeout=1.0)
                        while True:
                            try:
                                com.recv(timeout=1.0)
                                break
                            except RuntimeError as e:
                                # 期待しないデータは読み捨てる
                                log.warning(f"Unexpected data recv: {e}")
                    log.info(f"----- connected {self}")
                    break
                except (ConnectionError, timeout, TimeoutError) as e:
                    self._sock.close()
                    count += 1
                    if (retry is None) or (count < retry):
                        log.info(f"{self} retry {count}/{retry} by {type(e)}")
                        sleep(self.RETRY_INTERVAL)
                        self._sock = socket(AF_INET, SOCK_STREAM)
                        continue
                    else:
                        log.warn(f"{self} refused by server by {type(e)}")
                        self._sock = None
                        # リトライ回数を超えたら例外出力
                        raise e

    def close(self):
        """ 接続完了
        """
        if self._sock:
            self._sock.close()
            log.info(f"----- close {self}")

            self._sock = None

            # log.debug(f"release lock {id(self)}")
            # self._lock.release()

    def check_server(self):
        """ サーバの起動確認
        """
        if self._sock:
            # 接続済みならば起動確認OK
            return True

        # 未接続ならばサーバポートのListen待ちを確認する
        try:
            # 個別IPに対する待ち受けを確認
            dummy_sock = socket(AF_INET, SOCK_STREAM)
            dummy_sock.bind((self._host, self._port))
            dummy_sock.close()
            # 全IPに対する待ち受けも確認する
            dummy_sock = socket(AF_INET, SOCK_STREAM)
            dummy_sock.bind(('0.0.0.0', self._port))
            dummy_sock.close()
        except socket_error as e:
            if e.errno == errno.EADDRINUSE:
                # バインド済みならば起動確認OK
                return True
        # 未バインドならば起動確認NG
        return False

    def __str__(self):
        try:
            sockname = self._sock.getsockname()
            peername = self._sock.getpeername()
            return f"<Connection: {sockname[0]}:{sockname[1]} -> {peername[0]}:{peername[1]}>"
        except Exception:
            # ホスト名が取得できない場合は未接続
            return f"<Connection: unconnected -> {self._host}:{self._port}>"

    # 読み取り専用変数の定義
    @property
    def sock(self) -> socket:
        return self._sock



# (c) Roxy Corp. 2020-
# Roxy AI Coordinator
from __future__ import annotations
from typing import Union
from pathlib import Path
from logging import getLogger
from os import environ, getpid
from mmap import mmap, ACCESS_READ, ACCESS_WRITE
from shutil import rmtree


log = getLogger(__name__)


class _RingBufferSet():
    def __init__(self):
        """ リング式バッファセット
        """
        self.list: list[InspectBuffer] = []
        self.index: int = 0

    def get_next(self) -> InspectBuffer:
        """ 次のバッファを取得
        """
        self.index = (self.index + 1) % len(self.list)
        return self.list[self.index]

    @property
    def buffer(self) -> InspectBuffer:
        """ 現在のバッファを取得
        """
        return self.list[self.index]


class InspectBuffer():
    _root_path = Path(environ.get('TEMP')) / 'roxyai_api_buffer' / str(getpid())

    # バッファ名に対応するリングバッファ辞書
    _name_dict: dict[str, _RingBufferSet] = {}

    @classmethod
    def clean(cls):
        """ Roxy AI のメモリマップドファイルを全て削除する
        Note:
            ロックされているファイルは無視
            起動時に1度だけ実行する想定
        """
        if cls._root_path.parent.exists():
            for path in cls._root_path.parent.iterdir():
                try:
                    if path.is_dir():
                        rmtree(str(path))
                    else:
                        path.unlink()
                except Exception as e:
                    log.info(
                        f'{cls.__name__}: failed to clear {path}: {e}'
                    )

    @classmethod
    def release_all(cls):
        """ 管理しているバッファを全て開放する
        """
        for name, lst in cls._name_dict.items():
            for i, buf in enumerate(lst.list):
                try:
                    buf._release()
                except Exception as e:
                    log.warning(
                        f'{cls.__name__}: failed to release buffer '
                        f'"{name}:{i}"\n{e}'
                    )
        cls._name_dict.clear()

    @classmethod
    def alloc_set(cls, set_name: str, set_size: int, buffer_size: int) -> InspectBuffer:
        """ バッファセットをメモリマップ上に取得する
        Args:
            set_name:       バッファセット名
            set_size:       セットで作成するバッファ数
            buffer_size:    バッファサイズ[bytes]
        Returns:
            InspectBuffer:   作成した先頭のバッファインスタンス
            None:            バッファの作成失敗
        """
        buf = None
        if set_size < 1:
            ValueError(f'{cls.__name__}: number must be greater than 0')

        buf_set = cls._name_dict.get(set_name)
        if buf_set:
            # 既に同名のバッファセットを生成済み
            if (buf_set.buffer.size != buffer_size):
                log.warning(
                    f'{cls.__name__}: {set_name}:{buf.id} size is different from {buffer_size}'
                )
            buf = buf_set.get_next()
        else:
            try:
                buf_set = _RingBufferSet()
                for index in range(set_size):
                    # バッファの確保
                    file_name = f'{set_name}_{index:03d}'
                    path = cls._create_temp_file(file_name, buffer_size)
                    # 書き込み可能でバッファを取得
                    buff = cls(path, read_only=False)
                    buf_set.list.append(buff)
                # バッファセットを登録
                cls._name_dict[set_name] = buf_set
                buf = buf_set.buffer
            except Exception as e:
                log.warning(
                    f'{cls.__name__}: falied to allocate buffer, '
                    f'name: "{set_name}", nubmer: {set_size}, '
                    f'size: {buffer_size:,d}\n{e}'
                )
        return buf

    @classmethod
    def get_next_buffer(cls, set_name) -> mmap:
        """ 次のバッファを取得する
        Args:
            set_name: バッファ名
        Returns:
            InspectBuffer: 次のバッファインスタンス
            None:          バッファがない
        """
        ret = None
        if cls._name_dict.get(set_name):
            lst = cls._name_dict[set_name]
            cls._name_index[set_name] = 0
            if len(lst) > 0:
                ret = lst.pop(0)
        return ret

    @classmethod
    def _create_temp_file(cls, name: str, size: int) -> Path:
        """ メモリマップドファイル用のファイルを生成する
        Args:
            name:   ファイル名
            size:   ファイルサイズ[bytes]
        Returns:
            Path:   ファイルパス
        """
        try:
            # 格納ディレクトリ作成
            cls._root_path.mkdir(parents=True, exist_ok=True)
            path = cls._root_path / name
            if path.exists():
                log.warning(f'{cls.__name__}: file already exists: {path}')
            # Windowsは空のファイルを利用できないので適当なファイルを作成
            with open(path, 'wb') as fd:
                fd.write(b'\0' * size)
                fd.flush()
        except Exception as e:
            log.warning(
                f'{cls.__name__}: failed to allocate buffer, '
                f'name: "{name}", size: {size:,d}\n{e}'
            )
            raise
        return path

    def __init__(self, path: Union[str, Path], read_only: bool = True):
        """ 検査用のバッファ管理
        Args:
            path:       メモリマップドファイルパス
            read_only:  読み取り専用か
        Notes:
            メモリマップドファイル上にバッファを確保する
        """
        self._path = Path(path).resolve()

        self._name = self._path.name[:-4]
        self._id = int(self._path.name[-3])

        self._fd = None
        self._mm = None
        self._size = None
        self._read_only = read_only

        # インスタンス生成時にバッファ獲得
        self._mmap()

    def _mmap(self) -> bool:
        """ バッファをメモリマップ上に取得する
        Returns:
            True:  獲得成功
            False: 獲得失敗
        """
        try:
            # ファイルオープン
            self._fd = open(self._path, 'r+b')
            if self._read_only:
                access = ACCESS_READ
            else:
                access = ACCESS_WRITE
            # メモリマップ割り当て(確保メモリサイズはファイルサイズ)
            self._mm = mmap(self._fd.fileno(), 0, access=access)
            self._size = self._mm.size()
        except Exception:
            log.warning(f'{self}: failed to allocate buffer')
            raise
        log.debug(f'{self}: allocated buffer')

    def _release(self):
        """ バッファを開放する
        """
        # メモリマップ開放
        try:
            if self._mm.closed:
                self._mm.close()
        except Exception as e:
            log.warning(f'{self}: failed to close mmap: {e}')
        self._mm = None
        # ファイルクローズ
        try:
            if self._fd and (not self._fd.closed):
                self._fd.close()
        except Exception as e:
            log.warning(f'{self}: failed to close mmap: {e}')
        self._fd = None

    def get_next(self) -> InspectBuffer:
        """ 次のバッファを取得する
        Returns:
            InspectBuffer: 次のバッファインスタンス
            None:          次のバッファがない
        Note:
            リングバッファのセットで確保した場合のみ利用可能
        """
        buf = None
        buf_set = self._name_dict.get(self._name)
        if buf_set:
            buf = buf_set.get_next()
        else:
            log.warning(f'{self}: not found buffer set {self._name}')
        return buf

    @property
    def closed(self) -> bool:
        """ メモリマップドファイル開放済みか
        """
        return (not self._mm)

    @property
    def mmap(self) -> mmap:
        """ メモリマップドファイル
        """
        return self._mm

    @property
    def size(self) -> int:
        """ バッファサイズ
        """
        return self._size

    @property
    def path(self) -> Path:
        """ メモリマップのファイルパス
        """
        return self._path

    def __str__(self) -> str:
        text = f'<{self.__class__.__name__} '
        text += f'{self._name}:{self._id} '
        text += 'RO> ' if self._read_only else 'RW> '
        text += f'({self._size:,d} bytes)'
        return text


if __name__ == '__main__':
    import numpy as np
    from timeit import timeit
    from shutil import copyfileobj
    InspectBuffer.clean()
    mm = InspectBuffer.alloc_set('test', 2, 512 * 1024)
    print(mm)
    dt = np.dtype('uint8')
    dt = dt.newbyteorder('<')
    dst_np = np.frombuffer(mm.mmap, dtype=dt, offset=0)
    src_np = dst_np.copy()
    m2 = mm.get_next()

    # コピースピードの確認
    def test_func1():
        # データサイズが小さい時（～500KBくらい）には最速
        dummy_np = dst_np.copy()

    def test_func2():
        # test_func3より若干遅い
        mm.mmap[:] = dst_np.data[:]

    def test_func3():
        # データサイズが大きい時（1MB～）には最速
        copyfileobj(mm.mmap, m2.mmap)

    def test_func4():
        m2.mmap[:] = mm.mmap[:]

    result1 = timeit(test_func1, number=10)
    print(f'np.copy(): {result1}')
    result2 = timeit(test_func2, number=10)
    print(f'np.copy(): {result2}')
    result3 = timeit(test_func3, number=10)
    print(f'np.copy(): {result3}')
    result4 = timeit(test_func4, number=10)
    print(f'np.copy(): {result4}')
    pass



# (c) Roxy Corp. 2020-
# Roxy AI Inspect-Server API
from __future__ import annotations
from typing import Any, Union
import os
from pathlib import Path
import json5 as json
import re
from termcolor import cprint
from shutil import copy2
from logging import getLogger

log = getLogger(__name__)


class BaseConfig():
    _path = None
    _items = {}

    ROOT_PATH = Path(os.environ.get('Roxy_AI', 'C:/RoxyAI')) / 'roxy-ai-runtime'
    DEFAULT_FOLDER = ROOT_PATH / 'config/default'
    DEFAULT_PATH = None

    def __init__(self, src: dict[str, Any]):
        self._items = {}
        self.update(src)

    def update(self, src: dict[str, Any]):
        """ 設定値の内容を辞書から更新
        Args:
            src (dict):     値を読み込む辞書
        """
        for key in src.keys():
            if key in dir(self):
                if (
                    # 定数や内部変数、メソッド、プロパティなら無視する
                    key[0] == "_"
                    or key.isupper()
                    or callable(getattr(self, key))
                    or key in ('path', 'items')
                ):
                    continue

                val = src[key]
                setattr(self, key, val)
                self._items[key] = val
                # 開発中のデバグ用
                # print(f'>>>>>>>>> self: {self}, attr: {key}, val: {val}')

    @classmethod
    def load(cls, path: Union[str, Path] = None) -> BaseConfig:
        """ 行コメントを含むJSONファイルの読み込み
        Args:
            path (str)      JSON形式の設定ファイルパス
        Returns:
            dict            読み込んだ設定ファイルの辞書
        """
        if path:
            path = Path(path).expanduser().resolve().absolute()
        else:
            if not cls.DEFAULT_PATH:
                raise NotImplementedError()
            path = Path(cls.DEFAULT_PATH)

        if not path.exists():
            log.warning(f'{cls.__name__}: not found configuration file: {path}')
            default_file = cls.DEFAULT_FOLDER / path.name
            if default_file.exists():
                try:
                    copy2(default_file.as_posix(), path)
                    cprint(
                        f'Created a config file "{path}" '
                        f'by coping it from "{default_file}".',
                        color='yellow'
                    )
                except Exception as e:
                    log.error(
                        f'{cls.__name__}: cannot create a config file '
                        f'because copy from {default_file} failed.\n'
                        f'{e}'
                    )
                    raise
            else:
                raise RuntimeError(
                    f'{cls.__name__}: not found configuration file: {path}'
                )
        try:
            data = path.read_text(encoding="utf_8")
            dic = json.loads(data)
            instance = cls(dic)
            instance._path = path
        except Exception as e:
            raise RuntimeError(
                f'failed to load configuration file: {path}\n{e}'
            )
        return instance

    @property
    def items(self) -> dict[str, Any]:
        """ 設定ファイル情報の一覧取得
        """
        return self._items

    @property
    def path(self) -> Path:
        """ 設定ファイルのパス
        """
        return self._path

    def dump_items(
        self, dumper, prefix: str, max_length: int = 255, default=True, exclusion=[]
    ):
        """ ユーザが設定したデータの一覧をログ出力
        """
        if default:
            # デフォルト値を含む場合
            base_attrs = dir(super())
            items = {
                attr: getattr(self, attr) for attr in dir(self)
                if (
                    attr not in base_attrs
                    and attr[:1].islower()
                    and not callable(getattr(self, attr))
                    and attr not in ('path', 'items')
                )
            }

        else:
            items = self._items

        for name, val in items.items():
            if name in exclusion:
                # 除外のキーは除く
                continue
            text = f'{prefix}{name:25}: {val}'
            if len(text) >= max_length - 3:
                text = text[:(max_length - 3)] + '...'
            dumper(text)

    def __str__(self) -> str:
        name = self.__class__.__name__.replace('Config', '')
        text = f'{name} : {self._path}'
        return text


class CommonConfig(BaseConfig):
    """ 共通設定ファイル
    """
    # 定数定義
    ROOT_PATH = BaseConfig.ROOT_PATH
    DEFAULT_PATH = ROOT_PATH / 'config/common.json'

    # デフォルト設定値
    product_top = str(ROOT_PATH / 'roxy-ai-runtime/fixed_model')

    @classmethod
    def load(cls, path: Union[str, Path] = None) -> CommonConfig:
        return super().load(path)


class ServerConfig(BaseConfig):
    """ 共通サーバ設定ファイル
    """
    # 定数定義
    ROOT_PATH = BaseConfig.ROOT_PATH
    DEFAULT_PATH = ROOT_PATH / 'config/server.json'

    def __init__(self, dic: dict[str, Any]):
        super().__init__(dic)
        for item, val in dic.items():
            if type(val) is list:
                hosts = []
                ports = []
                for v in val:
                    mr = re.match(r'(?P<host>(\d+\.\d+\.\d+\.\d+))\:(?P<port>(\d+))', v)
                    if mr:
                        hosts.append(mr['host'])
                        ports.append(int(mr['port']))
                setattr(self, item + '_host', hosts)
                setattr(self, item + '_port', ports)
            self._items[item] = val

    @classmethod
    def load(cls, path: Union[str, Path] = None) -> ServerConfig:
        return super().load(path)



# (c) Roxy Corp. 2020-, CONFIDENTIAL
# Roxy AI Inspect-Server communication definition
from __future__ import annotations
from typing import NamedTuple, Union
from enum import IntEnum

# フレームのヘッダ定義
SIGN_CODE = 0x6941
HEADER_SIZE = 8


class CommandStatus(IntEnum):
    """ コマンドステータスの列挙クラス
    """
    NONE = -1
    STS_REQUEST = 0x00
    STS_REPLY_ACK = 0x01
    ERR_INVALID_SIGNATURE = 0x0B
    ERR_FIALED_PARSE_DATA = 0x0C
    ERR_INVALID_DATA_SIZE = 0x0D
    ERR_UNKNOWN_EXCEPTION = 0x0E
    ERR_INVALID_COMMAND = 0x0F
    ERR_NOT_FOUND_PRODUCT = 0x11
    ERR_NOT_FOUND_MODEL = 0x12
    ERR_DENIED_PRODUCT = 0x13
    ERR_FAILED_LOAD_CONFIG = 0x14
    ERR_FAILED_INITIALIZE = 0x15
    ERR_FAILED_LOGGING = 0x16

    ERR_INVALID_MODEL_ID = 0x21
    ERR_INVALID_IMG_FORMAT = 0x22
    ERR_INVALID_IMG_DATA = 0x23
    ERR_INVALID_JSON_DATA = 0x24
    ERR_UNINITIALIZED = 0x25
    ERR_OVERLAP_INSPECT_ID = 0x26
    ERR_FAILED_INSPECT = 0x27
    ERR_NOT_FOUND_PROB = 0x28
    ERR_NOT_FOUND_IMAGE = 0x29

    ERR_NOT_FOUND_SERVER = 0x31
    ERR_NOT_FOUND_CAMERA = 0x32
    ERR_NOT_FOUNT_MONITOR = 0x33
    ERR_NOT_DEFINED_PROC = 0x34
    ERR_INVALID_PRODUCT_CFG = 0x35
    ERR_INVALID_SHOT_NUMBER = 0x36
    ERR_DISCONNECTED_SERVER = 0x37
    ERR_FAILED_INIT_SERVER = 0x39

    ERR_FAILED_START_CAMERA = 0x41
    ERR_NOT_READY_CAMERA = 0x42
    ERR_FAILED_TRIGGER = 0x43
    ERR_FAILED_SETUP_CAMERA = 0x44
    ERR_FAILED_STOP_CAMERA = 0x45

    ERR_CDNT_NOT_ACTIVATED = 0x50

    ERR_INVALID_TARGET_WORK = 0x61
    ERR_RESULT_NOT_FOUND = 0x62

    @property
    def is_ack(self) -> bool:
        """ コマンド状態が正常応答か
        """
        return (self.value == self.STS_REPLY_ACK)

    @property
    def is_reply(self) -> bool:
        """ コマンド状態が応答コマンドか
        """
        return (self.value not in (
            self.STS_REQUEST,
            self.NONE,
        ))

    @property
    def is_error_reply(self) -> bool:
        """ コマンド状態がエラーの応答コマンドか
        """
        return (self.value not in (
            self.NONE,
            self.STS_REQUEST,
            self.STS_REPLY_ACK
        ))

    def __str__(self) -> str:
        message = _STATUS_DESCRIPTIONS.get(self.value, 'Unknown Status')
        return f"{message} (0x{self.value:02X})"


_STATUS_DESCRIPTIONS = {
    CommandStatus.NONE: '<<None>>',
    CommandStatus.STS_REQUEST: 'Request',
    CommandStatus.STS_REPLY_ACK: 'ACK',
    CommandStatus.ERR_INVALID_SIGNATURE: 'ERR: Invalid command signature',
    CommandStatus.ERR_FIALED_PARSE_DATA: 'ERR: Failed to parse data content',
    CommandStatus.ERR_INVALID_DATA_SIZE: 'ERR: Invalid command data size',
    CommandStatus.ERR_UNKNOWN_EXCEPTION: 'ERR: Unknown exception',
    CommandStatus.ERR_INVALID_COMMAND: 'ERR: Unknown command',
    CommandStatus.ERR_NOT_FOUND_PRODUCT: 'ERR: Cannot find product folder',
    CommandStatus.ERR_NOT_FOUND_MODEL: 'ERR: Cannot find model data',
    CommandStatus.ERR_DENIED_PRODUCT: 'ERR: Denied open additional product',
    CommandStatus.ERR_FAILED_LOAD_CONFIG: 'ERR: Failed to load config file',
    CommandStatus.ERR_FAILED_INITIALIZE: 'ERR: Failed to initialize model',
    CommandStatus.ERR_FAILED_LOGGING: 'ERR: Failed to output inspection log',

    CommandStatus.ERR_INVALID_MODEL_ID: 'ERR: Invalid model id',
    CommandStatus.ERR_INVALID_IMG_FORMAT: 'ERR: Invalid image format id',
    CommandStatus.ERR_INVALID_IMG_DATA: 'ERR: Invalid image data',
    CommandStatus.ERR_INVALID_JSON_DATA: 'ERR: Invalid JSON data',
    CommandStatus.ERR_UNINITIALIZED: 'ERR: Uninitialized',
    CommandStatus.ERR_OVERLAP_INSPECT_ID: 'ERR: Overlapped inspect id',
    CommandStatus.ERR_FAILED_INSPECT: 'ERR: Failed to inspection',
    CommandStatus.ERR_NOT_FOUND_PROB: 'ERR: Cannot find probabilities list',
    CommandStatus.ERR_NOT_FOUND_IMAGE: 'ERR: Cannot find image data',

    CommandStatus.ERR_NOT_FOUND_SERVER: 'ERR: Cannot find Inspect-Sever',
    CommandStatus.ERR_NOT_FOUND_CAMERA: 'ERR: Cannot find camera device',
    CommandStatus.ERR_NOT_FOUNT_MONITOR: 'ERR: Cannot find Monitor',
    CommandStatus.ERR_NOT_DEFINED_PROC: 'ERR: Not defined inspect procedure',
    CommandStatus.ERR_INVALID_PRODUCT_CFG: 'ERR: Inivalid product config file',
    CommandStatus.ERR_INVALID_SHOT_NUMBER: 'ERR: Inivalid shot number',
    CommandStatus.ERR_DISCONNECTED_SERVER: 'ERR: Server is disconnected',
    CommandStatus.ERR_FAILED_INIT_SERVER: 'ERR: Failed to initialize server',

    CommandStatus.ERR_FAILED_START_CAMERA: 'ERR: Failed to start camera',
    CommandStatus.ERR_NOT_READY_CAMERA: 'ERR: Camera is not ready',
    CommandStatus.ERR_FAILED_TRIGGER: 'ERR: Software trigger failed',
    CommandStatus.ERR_FAILED_SETUP_CAMERA: 'ERR: Failed to setup camera',
    CommandStatus.ERR_FAILED_STOP_CAMERA: 'ERR: Failed to stop camera',

    CommandStatus.ERR_CDNT_NOT_ACTIVATED: 'ERR: Coordinator is not activated',

    CommandStatus.ERR_RESULT_NOT_FOUND: 'ERR: Cannot find results',
    CommandStatus.ERR_INVALID_TARGET_WORK: 'ERR: Invalid target work',
}


# コマンドの定義
class CommandCode(IntEnum):
    """ コマンド番号の列挙クラス
    """
    # コマンドの定義
    ECHO = 0x10
    INITIALIZE = 0x11
    TERMINATE = 0x12
    INSPECT = 0x13
    GET_PROBABILITIES = 0x1A
    GET_IMAGE = 0x1B

    def __str__(self) -> str:
        message = _COM_NAME_LIST.get(self.value, 'Unknown Command Code')
        return f"{message}(0x{self.value:02X})"


_COM_NAME_LIST = {
    CommandCode.ECHO: 'Echo',
    CommandCode.INITIALIZE: 'Initialize',
    CommandCode.TERMINATE: 'Terminate',
    CommandCode.INSPECT: 'Inspect',
    CommandCode.GET_PROBABILITIES: 'GetProbabilities',
    CommandCode.GET_IMAGE: 'GetImage',
}


class Judgment(IntEnum):
    """ 判定結果の列挙クラス
    """
    NONE = 0x00         # 未判定
    OK = 0x1
    NOK = 0x2
    GRAY = 0x3
    UNK = 0x3           # deprecated
    FAILED = 0xFF

    def __and__(self, other) -> Judgment:
        """ 論理演算 OK < GRAY < NOK < FAILED の順で優先
        """
        priority = (self.FAILED, self.NOK, self.GRAY, self.OK, self.NONE)
        try:
            hi = min(priority.index(self), priority.index(other))
        except ValueError:
            return self.NONE
        return priority[hi]

    def __or__(self, other) -> Judgment:
        """ 論理演算 OK > GRAY > NOK > FAILED の順で優先
        """
        priority = (self.NONE, self.FAILED, self.NOK, self.GRAY, self.OK)
        try:
            hi = max(priority.index(self), priority.index(other))
        except ValueError:
            return self.NONE
        return priority[hi]

    def __str__(self) -> str:
        return f"{self.name:3s}(0x{self.value:02X})"


class JudgmentCount(NamedTuple):
    """ 判定結果のカウントクラス
    """
    ok: int = 0
    nok: int = 0
    gray: int = 0
    failed: int = 0

    @property
    def total(self) -> int:
        """ カウントの合計
        """
        return sum(self)


class JudgmentCounter():
    """ 判定結果のカウンタークラス
    """
    def __init__(self):
        self.reset()

    def reset(self):
        self.ok = 0
        self.nok = 0
        self.gray = 0
        self.failed = 0

    def set(self, judge: Judgment):
        if judge == Judgment.OK:
            self.ok += 1
        elif judge == Judgment.NOK:
            self.nok += 1
        elif judge == Judgment.GRAY:
            self.gray += 1
        elif judge == Judgment.FAILED:
            self.failed += 1

    def get_count(self) -> JudgmentCount:
        return JudgmentCount(self.ok, self.nok, self.gray, self.failed)

    def __str__(self) -> str:
        output = f'OK: {self.ok:,d}, NOK: {self.nok:,d}, GRAY: {self.gray:,d}'
        if self.failed:
            output += f', FAILED: {self.failed}'
        return output


class ProbabilityType(IntEnum):
    """ エリア確信度の型
    """
    SIMPLE = 0x01
    LABELED = 0x02

    def __str__(self) -> str:
        return f"{self.name:s}(0x{self.value:02X})"


class Probability():
    """ 判定結果の確信度クラス
    """
    def __init__(self, x1: int, y1: int, x2: int, y2: int, typ: int, prob: float):
        self.x1 = x1
        self.y1 = y1
        self.x2 = x2
        self.y2 = y2
        self.typ = typ
        self.prob = prob
        self.judge = Judgment(typ)

    def __str__(self) -> str:
        return (
            f'{self.judge.name:3s} '
            f'({self.x1:4d}, {self.y1:4d})-({self.x2:4d}, {self.y2:4d}) '
            f'{self.prob:6f}'
        )


class LabeledProbability(Probability):
    """ 不良種別情報付きの判定結果の確信度クラス
    """
    def __init__(self, x1, y1, x2, y2, typ, label, prob):
        super().__init__(x1, y1, x2, y2, typ, prob)
        self.label = label

    def __str__(self) -> str:
        return super().__str__() + f', label={self.label}'


class InspectResult():
    def __init__(
        self,
        model_name: str,
        group_name: str,
        judge: Judgment,
        prob_list: list[Union[Probability, LabeledProbability]],
        product_name: str = '',
    ):
        """ AI検査結果詳細
        """
        self.product_name = product_name
        self.model_name = model_name
        self.group_name = group_name
        self.judge = judge
        self.prob_list = prob_list

    @staticmethod
    def merge(all_results: list) -> InspectResult:
        """ AI検査結果詳細の統合
        Note:
            記録用に複数の検査結果をまとめる
        """
        model_name = '+'.join(sorted(list(set(r.model_name for r in all_results))))
        group_name = '+'.join(sorted(list(set(r.group_name for r in all_results))))
        judge = Judgment.NONE
        prob_list = []
        for result in all_results:
            judge &= result.judge
            prob_list.extend(result.prob_list)
        merged = InspectResult(
            model_name=model_name,
            group_name=group_name,
            judge=judge,
            prob_list=prob_list
        )

        return merged

    @property
    def model_group(self) -> str:
        """ モデル名（グループ名）の取得
        Note:
            グループがデフォルトの場合は表示しない
        """
        if self.group_name:
            return f'{self.model_name}({self.group_name})'
        else:
            return self.model_name



# (c) Roxy Corp. 2020-
# Roxy AI Analyze-Server API
from __future__ import annotations
import struct

from .connection import Connection
from .com_definition import (
    Probability,
    LabeledProbability,
    CommandCode,
    ProbabilityType,
)
from .com_base import BaseCommand


class GetProbabilitiesCommand(BaseCommand):

    _CODE = CommandCode.GET_PROBABILITIES
    PROB_SIZE = 13
    LABELED_PROB_SIZE = 14
    PROB_OFFSET = 3

    # ログの詳細出力
    verbose = False

    def __init__(
        self,
        inspect_id: int,
        connection: Connection = None,
        logging: bool = True,
    ):
        """ GetProbabilities コマンド制御
        Args:
            inspect_id:     取得対象の検査番号
            connection:     通信対象のTCP接続
            logging:        送受信時のログ出力フラグ
        """
        super().__init__(connection=connection, logging=logging)
        self.inspect_id = inspect_id
        self.data = struct.pack(f'< Q', inspect_id)

    def _decode_reply(self, reply: bytes):
        """ Inspect コマンドの応答内容確認

        Args:
            reply (bytes):      受信応答データ（ヘッダ以外）
        """
        prob_type, prob_len = struct.unpack(
            '< B H',
            reply[:self.PROB_OFFSET]
        )
        prob_data = reply[self.PROB_OFFSET:]

        # 受信データの格納
        self.prob_type = ProbabilityType(prob_type)
        self.prob_size = prob_len
        self.prob_list = []
        if prob_type == ProbabilityType.LABELED:
            prob_size = self.LABELED_PROB_SIZE
        else:
            prob_size = self.PROB_SIZE
        for offset in range(0, prob_len * prob_size, prob_size):
            if prob_type == ProbabilityType.LABELED:
                x1, y1, x2, y2, typ, label, prob = struct.unpack(
                    '< H H H H B B f',
                    prob_data[offset:offset + prob_size]
                )
                prob = LabeledProbability(x1, y1, x2, y2, typ, label, prob)
            else:
                x1, y1, x2, y2, typ, prob = struct.unpack(
                    '< H H H H B f',
                    prob_data[offset:offset + prob_size]
                )
                prob = Probability(x1, y1, x2, y2, typ, prob)
            self.prob_list.append(prob)

    def __str__(self):
        string = (
            f'{super().__str__()} '
            f'InspectID: {self.inspect_id}(=0x{self.inspect_id:#016X}) -> '
        )
        if self.is_received_ack:
            string += (
                f'ProbabilityList: {self.prob_size} items '
                f'({self.process_time} ms)'
            )
            if self.verbose:
                for prob in self.prob_list:
                    string += '\n    ' + str(prob)
        return string



# (c) Roxy Corp. 2020-
# Roxy AI Coordinator
from __future__ import annotations
from typing import Optional, Tuple, Union, NamedTuple
from enum import IntEnum
import numpy as np
from mmap import mmap
import struct
import cv2
import traceback
from pathlib import Path
from io import BytesIO
from urllib.parse import quote, urlparse, parse_qs, urlencode

from .inspect_buffer import InspectBuffer

from logging import getLogger
log = getLogger(__name__)


class ImageRect(NamedTuple):
    """ 領域指定用NamedTupleクラス
    """
    x1: int = 0
    y1: int = 0
    x2: int = 0
    y2: int = 0

    def __bool__(self) -> bool:
        """ 面積があればTrueを返す
        """
        return (self.x1 - self.x2) and (self.y1 - self.y2)


class ImageConvertionFailed(RuntimeWarning):
    """ 画像変換エラー
    """
    def __init__(self, src_fmt, dst_fmt, order, exception):
        if type(src_fmt) is int:
            src_fmt = ImageFormat(src_fmt)
        self.src_fmt = src_fmt

        if type(dst_fmt) is int:
            dst_fmt = ImageFormat(dst_fmt)
        self.dst_fmt = dst_fmt

        if type(order) is int:
            order = ColorOrder(order)
        self.order = order

        self.exception = exception

    def __str__(self):
        return (
            f'image conversion failed '
            f'from [{self.src_fmt.name}] '
            f'to [{self.dst_fmt.name}] '
            f'color order: {self.order.name}'
        )


class ImageFormat(IntEnum):
    """ イメージフォーマット値の定義クラス
    """
    NONE = 0x00
    RAW = 0x01
    JPEG = 0x02
    PNG = 0x03
    BMP = 0x04
    # URL形式の外部データ場所指定
    URL = 0xFE
    # 内部利用のフォーマット形式
    NDARRAY = 0xFF

    RAW_HADER_SIZE = 5

    def __str__(self):
        return f'{self.name}(0x{self.value:02X})'

    @property
    def suffix(self) -> str:
        return IMG_FORMAT_SUF(self.value)

    @staticmethod
    def all_suffix(header: str = None) -> tuple[str]:
        if header is None:
            return IMG_SUFFIX_LIST
        return tuple(header + s for s in IMG_SUFFIX_LIST)

    @classmethod
    def from_suffix(cls, suffix: str) -> ImageFormat:
        """ 拡張子からフォーマット値を取得
        Args:
            suffix (str)    拡張子文字列
        Returns:
            ImageFormat:    拡張子に対応するフォーマット値
            None:           拡張子が不正
        """
        suf = suffix.lower()
        if suf and suf[0] != '.':
            suf = '.' + suf
        val = IMG_SUFFIX_FMT.get(suf)
        if val:
            return cls(val)
        return None

    @classmethod
    def check_header(cls, data: memoryview):
        """ 画像データのヘッダからフォーマットを取得
        Args:
            data:   画像データのヘッダ
        Returns:
            ImageFormat:    解析成功
            None:           解析失敗
        Note:
            data は bytes でも良い。
        """
        size = len(data)

        # PNGフォーマット(http://www.w3.org/TR/PNG/)の読み取り
        # ※ 古いPNGフォーマットは非対応
        if (size >= 8) and (data[:8] == b'\x89PNG\r\n\x1A\n'):
            return cls.PNG
        if (size > 2) and (data[:2] == b'BM'):
            return cls.BMP
        if (size > 4) and (data[:2] == b'\xFF\xD8'):
            return cls.JPEG
        if size > cls.RAW_HADER_SIZE:
            # RAWフォーマットとサイズ一致ならRAWと判定
            width, height, order = struct.unpack('< H H B', data[:5])
            depth = ColorOrder(order).depth
            if size == (cls.RAW_HADER_SIZE + (width * height * depth)):
                return cls.RAW
        return None

    @staticmethod
    def set_raw_image(buffer: memoryview, width: int, height: int, order: ColorOrder) -> np.ndarray:
        """ バッファにRAWフォーマットのヘッダを設定する
        Args:
            buffer:     データバッファ
            width:      イメージの幅
            height:     イメージの高さ
            orger:      カラーフォーマット
        Returns:
            None:       設定失敗
            np.ndarray: 画像データの ndarray
        """
        data = None
        if len(buffer) > 5:
            buffer[0:2] = struct.pack('<H', width)
            buffer[2:4] = struct.pack('<H', height)
            buffer[4] = order.value
            # データ領域のndarrayへの変換
            dt = np.dtype('uint8')
            dt = dt.newbyteorder('<')
            data: np.ndarray = np.frombuffer(
                buffer, dtype=dt, offset=ImageFormat.RAW_HADER_SIZE
            )
            data = data.reshape(height, width, order.depth)
        return data


class ColorOrder(IntEnum):
    """ Rawイメージの色データバイト順の定義クラス
    """
    GRAY = 1
    BGR = 2
    RGB = 3
    BGRA = 4

    BAYER_RG = 10
    BAYER_BG = 11

    @property
    def depth(self):
        """ 色順序による色深度（ピクセルあたりバイト数）
        """
        if self == ColorOrder.GRAY:
            ret = 1
        elif self in (
            ColorOrder.BGR, ColorOrder.RGB,
            ColorOrder.BAYER_BG, ColorOrder.BAYER_RG,
        ):
            ret = 3
        elif self == ColorOrder.BGRA:
            ret = 4
        return ret

    def __str__(self):
        return f'{self.name}({self.value:d})'


# フォーマット変換の定義
IMG_FORMAT_SUF = {
    ImageFormat.RAW: '.bin',
    ImageFormat.JPEG: '.jpg',
    ImageFormat.PNG: '.png',
    ImageFormat.BMP: '.bmp',
}

IMG_SUFFIX_LIST = tuple(IMG_FORMAT_SUF.values())

IMG_SUFFIX_FMT = {
    **{v: ImageFormat(k) for k, v in IMG_FORMAT_SUF.items()},
    '.jpeg': ImageFormat(ImageFormat.JPEG),
}

# CVの変換用パラメータ定義
__JPEG_PARAMS = (cv2.IMWRITE_JPEG_QUALITY, 95)
__PNG_PARAMS = (cv2.IMWRITE_PNG_COMPRESSION, 1)


def _bin_to_ndarary(src) -> tuple[np.ndarray, ColorOrder]:
    # 構造情報の設定
    width, height, order = struct.unpack('< H H B', src[:5])
    dt = np.dtype('uint8')
    dt = dt.newbyteorder('<')
    dst: np.ndarray = np.frombuffer(src, dtype=dt, offset=5)
    depth = ColorOrder(order).depth
    if depth == 1:
        dst = dst.reshape((height, width))
    else:
        dst = dst.reshape((height, width, depth))
    return dst, ColorOrder(order)


def _image_to_ndarary(src) -> tuple[np.ndarray, ColorOrder]:
    dt = np.dtype('uint8')
    dt = dt.newbyteorder('<')
    buf = np.frombuffer(src, dtype=dt, offset=0)
    dst: np.ndarray = cv2.imdecode(buf, cv2.IMREAD_IGNORE_ORIENTATION | cv2.IMREAD_UNCHANGED)
    if len(dst.shape) == 2:
        order = ColorOrder.GRAY
    elif dst.shape[2] == 3:
        # CVは24bitをBGRの色順で保持
        order = ColorOrder.BGR
    elif dst.shape[2] == 4:
        # CVは32bitをBGRAの色順で保持
        order = ColorOrder.BGRA
    return dst, ColorOrder(order)


def _ndarary_to_bin(src: np.ndarray, order: ColorOrder) -> bytes:
    shape = (src.shape[1], src.shape[0], order)
    dst = struct.pack('< H H B', *shape)
    dst += src.tobytes()
    return dst


def _ndarary_to_jpeg(src, order: ColorOrder) -> bytes:
    if order == ColorOrder.RGB:
        # CV2の色順に変換
        src = cv2.cvtColor(src, cv2.COLOR_RGB2BGR)
    ret, dst = cv2.imencode('.jpg', src, params=__JPEG_PARAMS)
    if not ret:
        RuntimeWarning('imencode for jpg failed')
    return bytes(np.frombuffer(dst, dtype=np.uint8))


def _ndarary_to_png(src, order: ColorOrder) -> bytes:
    if order == ColorOrder.RGB:
        # CV2の色順に変換
        src = cv2.cvtColor(src, cv2.COLOR_RGB2BGR)
    ret, dst = cv2.imencode('.png', src, params=__PNG_PARAMS)
    if not ret:
        RuntimeWarning('imencode for png failed')
    return bytes(np.frombuffer(dst, dtype=np.uint8))


def _ndarary_to_bmp(src, order: ColorOrder) -> bytes:
    if order == ColorOrder.RGB:
        # CV2の色順に変換
        src = cv2.cvtColor(src, cv2.COLOR_RGB2BGR)
    ret, dst = cv2.imencode('.bmp', src)
    if not ret:
        RuntimeWarning('imencode for bmp failed')
    return bytes(np.frombuffer(dst, dtype=np.uint8))


class ImageBuffer(BytesIO):
    name = 'Buffer'

    def __init__(self, *args, **kwargs):
        """ 分割受信用有りの画像専用バッファ
        """
        super().__init__(*args, **kwargs)

    def close(self, release: bool = None):
        # 通常の書き込み完了時にはクローズさせない
        if release:
            super().close()
        return


class InspectImage():
    """ 検査に利用するイメージデータのキャッシュ機能付き管理クラス
    """
    DECODER = {
        ImageFormat.RAW: _bin_to_ndarary,
        ImageFormat.JPEG: _image_to_ndarary,
        ImageFormat.PNG: _image_to_ndarary,
        ImageFormat.BMP: _image_to_ndarary,
    }

    ENCODER = {
        ImageFormat.RAW: _ndarary_to_bin,
        ImageFormat.JPEG: _ndarary_to_jpeg,
        ImageFormat.PNG: _ndarary_to_png,
        ImageFormat.BMP: _ndarary_to_bmp,
    }

    def __init__(
        self,
        fmt: ImageFormat,
        data: Union[bytes, np.ndarray],
        path: Optional[Path] = None,
        order: ColorOrder = None,
        name: str = '',
    ):
        """ 検査用の画像データ管理
        Args:
            fmt     data で渡す画像のフォーマット
                    不明は場合は None で指定
            data:   画像データ
            path    読み込んだ元ファイルのパス
            order:  画像の色データバイト順
            name:   画像の識別名称
        """
        self._org_fmt = ImageFormat(fmt)
        self._data = {self._org_fmt: data}

        self._path = Path(path) if path else None
        self._name = str(name)
        if not name:
            # 名前の指定が無ければ自動で定義
            path_name = None
            try:
                path_name = Path(path).name
            except Exception:
                pass
            if path_name:
                # 名称が未指定でパスがある場合はファイル名を設定
                self._name = path_name
            else:
                self._name = f'image[{self._org_fmt.name}]'
        self._keep_count = 1

        # カラーオーダーを設定
        if order is None:
            if fmt == ImageFormat.NDARRAY:
                if len(data.shape) == 2:
                    order = ColorOrder.GRAY
                elif data.shape[2] == 3:
                    # CVは24bitをBGRの色順で保持
                    order = ColorOrder.BGR
                elif data.shape[2] == 4:
                    # CVは32bitをBGRAの色順で保持
                    order = ColorOrder.BGRA
        self._order = order
        # キャッシュ管理情報
        self._parent: InspectImage = None
        self._children: list[InspectImage] = []
        self._zoom: float = None
        self._clip: ImageRect = None
        # バッファ管理情報
        self._src_buffer: BytesIO = None
        self._src_view: memoryview = None
        self._src_size = 0
        self._mmap_path = None

    @classmethod
    def get_empty(
        cls,
        width: int = 1,
        height: int = 1,
        color: Union[tuple[int, int, int], tuple[int, int, int, int], int] = (0, 0, 0)
    ) -> InspectImage:
        """ 空のイメージを取得する
        Args:
            width:      画像の幅ピクセル数
            height:     画像の高さピクセル数
            colose:     塗りつぶし色
                (int, int, int):         RGB画像として作成(3ch)
                (int, int, int, int):    RGBA画像として作成(4ch)
                int:                     Gray画像として作成(1ch)
        """
        if (type(color) is int):
            order = ColorOrder.GRAY
        elif len(color) == 1:
            color = color[0]
            order = ColorOrder.GRAY
        elif len(color) == 3:
            order = ColorOrder.RGB
        elif len(color) == 4:
            order = ColorOrder.BGRA
        else:
            raise TypeError(f'{cls.__name__}: invalid color parameter {color}')

        if order == ColorOrder.GRAY:
            data = np.full((height, width), color, np.uint8)
        else:
            data = np.full((height, width, order.depth), color, np.uint8)
        instance = InspectImage(ImageFormat.NDARRAY, data, order=order)
        return instance

    @classmethod
    def from_buffer(
        cls,
        buffer: Union[BytesIO, mmap, memoryview],
        name: str = None,
        path: Path = None,
        width: int = None,
        height: int = None,
        order: ColorOrder = None,
    ) -> InspectImage:
        """ バッファから画像を取得する
        Args:
            buffer:     データバッファ
            name:       ファイル名（パス指定）
            path:       ファイルパス
            width:      画像の幅ピクセル数
            height:     画像の高さピクセル数
            order:      画像の色データバイト順
        Notes:
            バッファへの読み込みが完了していない場合は
            バッファへの書き込み後に flush_buffer する必要がある。
        """
        instance = None

        if isinstance(buffer, BytesIO):
            completed = None
            with buffer.getbuffer() as view:
                completed = cls.check_data(view)
            if completed:
                # 読み込み完了済みならば初期化
                view = buffer.getbuffer()
                fmt = ImageFormat.check_header(view)
                instance = cls(fmt, view, name=name)
                instance._src_buffer = buffer
                instance._src_size = len(view)
                instance._src_view = view
        else:
            try:
                # 画像フォーマットと読込状態の確認
                fmt = cls.check_data(buffer)
                if fmt:
                    # 読み込み完了済みならば初期化
                    instance = cls(fmt, buffer, name=name, path=path)
                    instance._src_buffer = None
                    instance._src_size = len(buffer)
                    instance._src_view = buffer
                    if isinstance(buffer, mmap):
                        # メモリマップドファイルならパスを設定
                        instance._mmap_path = path
            except Exception as e:
                log.warning(
                    f'{cls.__name__}: failed to create an image from buffer\n{e}'
                )

        return instance

    @classmethod
    def from_url(cls, url: Union[str, bytes]) -> InspectImage:
        """ URLから画像を取得する
        Args:
            url:    URL
        """
        instance = None
        if type(url) is bytes:
            try:
                # URLデータの解析
                url = url.decode('utf-8', errors='ignore')
            except Exception as e:
                log.warning(f'{cls.__name__} url decode error: {e}')
        try:
            url_info = urlparse(url)
            url_prams = parse_qs(url_info.query)
        except Exception as e:
            log.warning(f'{cls.__name__} url parse error: {url}\n{e}')
            url_info = None
            url_prams = {}

        # URLパラメータを取得
        try:
            width = int(url_prams.get('width', 0))
            height = int(url_prams.get('height', 0))
            order_param = url_prams.get('order')
            order = ColorOrder(order_param)
            clip_x1 = int(url_prams.get('clip_x1', 0))
            clip_y1 = int(url_prams.get('clip_y1', 0))
            clip_x2 = int(url_prams.get('clip_x2', 0))
            clip_y2 = int(url_prams.get('clip_y2', 0))
            clip_area = ImageRect(clip_x1, clip_y1, clip_x2, clip_y2)
        except Exception as e:
            log.warning(f'{cls.__name__} url parameters error: {url_prams}\n{e}')
            url_info = None

        if url_info:
            if url_info.scheme == 'file':
                # ファイルからデータの読み込み
                instance = cls.load(url_info.path)
            elif url_info.scheme == 'mmap':
                # メモリマップドファイルから画像データを取得
                buf = InspectBuffer(url_info.path)
                if None not in (width, height, order):
                    # サイズが全てパラメーター指定されている場合は ndarray として読込
                    data = ImageFormat.set_raw_image(buf.mmap, width, height, order)
                    instance = InspectImage(
                        ImageFormat.NDARRAY,
                        data=data,
                        path=url_info.path,
                        order=order,
                    )
                    pass
                else:
                    # サイズ指定が無い場合はRAWなどの画像フォーマットとして読込
                    instance = cls.from_buffer(
                        buf.mmap,
                        path=url_info.path,
                    )

            if clip_area:
                # 画像データのクリップ処理
                instance = instance.get_clipped(clip_area)

        return instance

    @classmethod
    def load(cls, path) -> InspectImage:
        """ ファイルから画像を読み込む
        """
        instance = None
        imgfile = Path(path)
        if imgfile.exists():
            data = imgfile.read_bytes()
            fmt = cls.check_data(data)
            if fmt:
                instance = cls(fmt, data, path=imgfile)
        return instance

    def save(self, path: Union[str, Path]) -> bool:
        """ ファイルに画像を書き込む
        Returns:
            True:   書き込み成功
            False:  書き込み失敗
        Note:
            ファイル名の拡張子で画像フォーマットを決定する
        """
        imgfile = Path(path)
        fmt = IMG_SUFFIX_FMT.get(imgfile.suffix)
        if fmt:
            data = self.get_image(fmt)
            if data:
                imgfile.write_bytes(data)
                if not self._path:
                    # ファイルパスが未登録ならばパス登録
                    self._path = imgfile
                return True
        return False

    def get_tf_image(self) -> np.ndarray:
        """ TensorFlow用のデータ取得
        """
        img = self.get_image()
        if img.size > 0:
            # 画像が存在する場合のみ変換実施
            # ndarrayがTF形式RGB(24bit)でなければ変換
            if self._order == ColorOrder.GRAY:
                img = cv2.cvtColor(img, cv2.COLOR_GRAY2BGR)
            elif self._order == ColorOrder.BGR:
                img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
            elif self._order == ColorOrder.BGRA:
                img = cv2.cvtColor(img, cv2.COLOR_BGRA2RGB)
            elif self._order == ColorOrder.BAYER_BG:
                img = cv2.cvtColor(img, cv2.COLOR_BAYER_BG2RGB)
            elif self._order == ColorOrder.BAYER_RG:
                img = cv2.cvtColor(img, cv2.COLOR_BAYER_RG2RGB)
        return img

    def get_image(
        self, fmt: ImageFormat = ImageFormat.NDARRAY,
    ) -> Union[np.ndarray, bytes, None]:
        """ 指定フォーマトの画像データを取得
        """
        img = self._data.get(fmt)
        if img is not None:
            # 対象フォーマットがキャッシュにあればそれを返す
            return img
        # 無ければ画像変換してそれを返す
        self.__convert(fmt)
        img = self._data.get(fmt)
        return img

    def get_size(self, fmt: ImageFormat = None) -> int:
        """ 指定フォーマットのデータサイズを取得
        """
        fmt = ImageFormat(fmt)
        img = self.get_image(fmt)
        if img is None:
            return None
        if fmt == ImageFormat.NDARRAY:
            return img.size
        else:
            return len(img)

    def __convert(self, fmt: ImageFormat = ImageFormat.NDARRAY):
        """ 画像フォーマット変換
        """
        if self._data.get(fmt) is not None:
            # 変換済みなら何もしない
            return

        if self._order is None:
            # 基準データ(ndarray)の作成
            src = self._data.get(self._org_fmt)
            try:
                if (src is None) and self._src_buffer:
                    # バッファからの読み込み途中の場合はビューから変換
                    self._src_buffer.flush()
                    with self._src_buffer.getbuffer() as view:
                        fmt = ImageFormat.check_header(view)
                        converter = self.DECODER[fmt]
                        dst, order = converter(view)
                else:
                    converter = self.DECODER[self._org_fmt]
                    dst, order = converter(src)
            except Exception as e:
                log.warning(traceback.format_exc())
                raise ImageConvertionFailed(self._org_fmt, ImageFormat.NDARRAY, None, e)
            self._data[ImageFormat.NDARRAY] = dst
            self._order = order

        if fmt != ImageFormat.NDARRAY:
            # 基準データ(ndarray)から変換先の画像データの作成
            src = self._data[ImageFormat.NDARRAY]
            converter = self.ENCODER[fmt]
            try:
                dst = converter(src, self._order)
            except Exception as e:
                log.warning(traceback.format_exc())
                raise ImageConvertionFailed(ImageFormat.NDARRAY, fmt, self._order, e)
            self._data[fmt] = dst

    def keep(self):
        """ 画像データを保持する(TBD)
        """
        self._keep_count += 1

    def release(self):
        """ 画像データを全て破棄する(TBD)
        """
        self._keep_count -= 1
        if self._keep_count > 0:
            # まだ保持している処理があるので破棄しない
            return
        self._data.clear()
        self._org_fmt = None
        if self._fnc_buf_release:
            # バッファの解放コールバックの呼び出し
            try:
                self._fnc_buf_release()
            except Exception as e:
                log.warning(
                    f'{e}: in callback for release image buffer: '
                    f'{self._fnc_buf_release()} '
                    f'<{self._name}>'
                )
            self._fnc_buf_release = None

    def get_clipped(
        self,
        area: ImageRect,
    ) -> InspectImage:
        """ 領域をクリップしたイメージオブジェクトを返す
        """
        area = ImageRect(*area)
        # キャッシュ探索
        che = [i for i in self._children if i._clip == area]
        if che:
            return che[0]
        # 画像生成
        org = self.get_image()
        buf = org[area.y1:area.y2, area.x1:area.x2]
        img = InspectImage(ImageFormat.NDARRAY, buf)
        # キャッシュ情報登録
        img._parent = self
        img._clip = area
        img._mmap_path = self._mmap_path
        self._children.append(img)
        return img

    def get_resized(self, width: int, height: int) -> InspectImage:
        """ 画像をリサイズしたイメージオブジェクトを返す
        """
        # キャッシュの探索
        target_shape = (height, width)
        che = [i for i in self._children if i.get_image().shape == target_shape]
        if che:
            return che[0]
        # 画像生成
        org = self.get_image()
        buf = cv2.resize(org, (width, height))
        img = InspectImage(ImageFormat.NDARRAY, buf)
        # キャッシュ情報登録
        img._parent = self
        self._children.append(img)
        return img

    def get_grayed(self) -> InspectImage:
        """ 画像をグレースケール化したイメージオブジェクトを返す
        """
        if self.order is ColorOrder.GRAY:
            # グレースケール画像ならばそのまま利用
            img = self
        else:
            # キャッシュの探索
            che = [i for i in self._children if i._order is ColorOrder.GRAY]
            if che:
                return che[0]
            # 画像生成
            org = self.get_image()
            if self._order is ColorOrder.RGB:
                cvt_code = cv2.COLOR_RGB2GRAY
            elif self._order is ColorOrder.BGR:
                cvt_code = cv2.COLOR_BGR2GRAY
            elif self._order is ColorOrder.BGRA:
                cvt_code = cv2.COLOR_BGRA2GRAY
            elif self._order is ColorOrder.BAYER_BG:
                cvt_code = cv2.COLOR_BAYER_BG2GRAY
            elif self._order is ColorOrder.BAYER_RG:
                cvt_code = cv2.COLOR_BAYER_RG2GRAY
            else:
                log.warning(f'invalid color order {self._order} in <{self._name}>')
                cvt_code = None
            try:
                buf = cv2.cvtColor(org, cvt_code)
                img = InspectImage(ImageFormat.NDARRAY, buf)
                # キャッシュ情報登録
                img._parent = self
                self._children.append(img)
            except Exception as e:
                log.warning(traceback.format_exc())
                raise ImageConvertionFailed(
                    self._org_fmt, ImageFormat.NDARRAY, ColorOrder.GRAY, e
                )
        return img

    def get_rgb_image(self) -> InspectImage:
        """ 画像をRGBカラー化したイメージオブジェクトを返す
        """
        if self.order is ColorOrder.RGB:
            # RGB画像ならばそのまま利用
            img = self
        else:
            # キャッシュの探索
            che = [i for i in self._children if i._order is ColorOrder.RGB]
            if che:
                return che[0]
            # 画像生成
            org = self.get_image()
            if self._order is ColorOrder.GRAY:
                cvt_code = cv2.COLOR_GRAY2RGB
            elif self._order is ColorOrder.BGR:
                cvt_code = cv2.COLOR_BGR2RGB
            elif self._order is ColorOrder.BGRA:
                cvt_code = cv2.COLOR_BGRA2RGB
            elif self._order is ColorOrder.BAYER_BG:
                cvt_code = cv2.COLOR_BAYER_BG2RGB
            elif self._order is ColorOrder.BAYER_RG:
                cvt_code = cv2.COLOR_BAYER_RG2RGB
            else:
                log.warning(f'invalid color order {self._order} in <{self._name}>')
                cvt_code = None
            try:
                buf = cv2.cvtColor(org, cvt_code)
                img = InspectImage(ImageFormat.NDARRAY, buf)
                # キャッシュ情報登録
                img._parent = self
                self._children.append(img)
            except Exception as e:
                log.warning(traceback.format_exc())
                raise ImageConvertionFailed(
                    self._org_fmt, ImageFormat.NDARRAY, ColorOrder.RGB, e
                )
        return img

    def get_bgr_image(self) -> InspectImage:
        """ 画像をRGBカラー化したイメージオブジェクトを返す
        """
        if self.order is ColorOrder.BGR:
            # BGR画像ならばそのまま利用
            img = self
        else:
            # キャッシュの探索
            che = [i for i in self._children if i._order is ColorOrder.BGR]
            if che:
                return che[0]
            # 画像生成
            org = self.get_image()
            if self._order is ColorOrder.GRAY:
                cvt_code = cv2.COLOR_GRAY2BGR
            elif self._order is ColorOrder.RGB:
                cvt_code = cv2.COLOR_RGB2BGR
            elif self._order is ColorOrder.BGRA:
                cvt_code = cv2.COLOR_BGRA2BGR
            elif self._order is ColorOrder.BAYER_BG:
                cvt_code = cv2.COLOR_BAYER_GB2BGR
            elif self._order is ColorOrder.BAYER_RG:
                cvt_code = cv2.COLOR_BAYER_RG2BGR
            else:
                log.warning(f'invalid color order {self._order} in <{self._name}>')
                cvt_code = None
            try:
                buf = cv2.cvtColor(org, cvt_code)
                img = InspectImage(ImageFormat.NDARRAY, buf)
                # キャッシュ情報登録
                img._parent = self
                self._children.append(img)
            except Exception as e:
                log.warning(traceback.format_exc())
                raise ImageConvertionFailed(
                    self._org_fmt, ImageFormat.NDARRAY, ColorOrder.BGR, e
                )
        return img

    def get_url(self) -> str:
        """ 画像のURLを返す
        """
        url = None
        if self._mmap_path:
            # メモリマップ画像の場合は対応するURLを生成
            url = quote(f'mmap://{self._mmap_path}')
            if self._org_fmt == ImageFormat.NDARRAY:
                # NDARRAY画像の場合は画像形状をパラメータに追加
                param_dict = {
                    'width': self.width,
                    'height': self.height,
                    'order': self._order,
                }
        elif self._path:
            # ファイル画像の場合はURLにファイルパスを設定
            url = quote(f'file://{self._path}')

        if url and self._clip:
            # クリップされている場合はクリップ情報を付加
            clip_dict = {
                'clip_x1': self._clip.x1,
                'clip_y1': self._clip.y1,
                'clip_x2': self._clip.x2,
                'clip_y2': self._clip.y2,
            }
            param_dict.update(clip_dict)

        if url and param_dict:
            # パラメータがあればURLに付加
            url += f'?{urlencode(param_dict)}'
        return url

    def show_window(self):
        """ 画像のウィンドウ表示（開発用）
        """
        img = self.get_image()
        cv2.imshow(str(self._path), img)
        cv2.waitKey(0)
        cv2.destroyAllWindows()

    def release_buffer(self):
        if self._src_view:
            # ビューデータを削除
            self._src_view.release()
            self._data.pop(self._org_fmt)
        if self._src_buffer:
            # バッファをクローズ
            self._src_buffer.close(release=True)
            self._src_buffer = None

    # 読み取り専用のパラメータアクセス定義
    @property
    def format(self) -> int:
        return self._org_fmt

    @property
    def data(self) -> int:
        return self._data

    @property
    def path(self) -> Path:
        return self._path

    @property
    def name(self) -> str:
        return self._name

    @property
    def parent(self) -> InspectImage:
        return self._parent

    @property
    def zoom(self) -> float:
        return self._zoom

    @property
    def order(self) -> ColorOrder:
        return self._order

    @property
    def width(self) -> int:
        """ 画像の幅のピクセル数
        """
        return self.get_image().shape[1]

    @property
    def height(self) -> int:
        """ 画像の幅のピクセル数
        """
        return self.get_image().shape[0]

    @property
    def buffer(self) -> BytesIO:
        return self._src_buffer

    @property
    def mmap_path(self) -> str:
        """ メモリマップドファイルへのパス名
        """
        return self._mmap_path

    @staticmethod
    def get_merged(images: Tuple[InspectImage]) -> InspectImage:
        """ 画像をカラー合成したイメージオブジェクトを返す
        Args:
            images:     グレースケール画像のリスト
        Return:
            合成画像
        """
        # 画像生成
        try:
            src_data = []
            for img in images:
                img_data = img.get_image()
                if img_data.dtype == np.uint16:
                    img_data = (img_data // 0x100).astype(np.uint8)
                src_data.append(img_data)
            buf = cv2.merge(src_data)
            new_name = '_'.join(img.name for img in images if img.name)
            img = InspectImage(ImageFormat.NDARRAY, buf, name=new_name)
        except Exception as e:
            imgs_str = ', '.join(str(img._name) for img in images)
            log.warning(f'Failed to merge images ({imgs_str})\n{e}')
            img = None
        return img

    @staticmethod
    def check_data(data) -> ImageFormat:
        """ バッファのデータサイズをチェック
        Args:
            data:       チェックするバイナリ
        Returns:
            ImageFormat.XXXXX:      全データ取得済みの画像フォーマット
            ImageFormat.None:       画像データ取得未完了
        Note:
            バッファ操作でメモリ確保を行わない
        """
        size = len(data)
        format = ImageFormat.check_header(data)
        length = 0

        # PNGフォーマット確認
        if (
            format == ImageFormat.PNG
            and (size >= 24)
            and (data[12:16] == b'IHDR')
        ):
            # PNGファイルサイズの推測（データの存在する範囲で計算）
            length = 8 + 25     # Signature + IHDR
            while length + 8 <= size:
                # 読み出すチャンクがまだ存在するならば
                l, = struct.unpack('> L', data[length:length + 4])
                ctype = data[length + 4:length + 8]
                length += int(l) + 12       # Lenght + ChunkType + CRC (12bytes)

                # 解析用のログ出力を行う場合(デバグ用コード)
                # from logging import DEBUG
                # if log.root.level <= DEBUG:
                #     # デバグ出力有りの場合
                #     if type(ctype) is memoryview:
                #         bdata = ctype.tobytes()
                #     else:
                #         bdata = ctype
                #     log.debug(f'    PNG chunk {bdata.decode()} [{length:,} bytes]')

                if ctype == b'IEND':
                    # 終端なので解析終了
                    break
            if ctype != b'IEND':
                # 終端チャンク未到達ならば足し合わせる
                length += 12

        # BMPフォーマット確認
        elif (
            format == ImageFormat.BMP
            and (size >= 6)
        ):
            # PNGファイルのデータ読み取り
            format = ImageFormat.BMP
            l, = struct.unpack('< L', data[2:6])
            length = int(l)

        # JPEGフォーマット確認(TBD)
        elif format == ImageFormat.JPEG:
            # 暫定でJPEGは常に読み取り完了済みとする
            length = size

        # RAWフォーマット確認
        elif format == ImageFormat.RAW:
            # RAWフォーマット判定時はデータサイズ完了を保証
            length = size

        # 受信サイズの方が大きい場合は読込完了とする
        return format

    def __str__(self):
        if self._name:
            text = f'"{self._name}"'
        else:
            text = 'Image'
        text += f'({self.width}, {self.height}): '
        cached = [
            f'{ImageFormat(fmt).name} '
            f'{len(self._data.get(fmt)):,d} bytes'
            for fmt in IMG_FORMAT_SUF
            if type(self._data.get(fmt)) is bytes
        ]
        if self._order:
            # 基準データ(ndarray)作成済み
            raw = self._data.get(ImageFormat.NDARRAY)
            cached.append(
                f'{ImageFormat.NDARRAY.name}-'
                f'{self._order.name} '
                f'{raw.size:,d} bytes'
            )
        text += ', '.join(cached)
        if self._path:
            try:
                path = Path(self._path).resolve().absolute().as_posix()
            except Exception:
                path = str(self._path)
            # ファイルに紐づいている場合
            text += f', path: {path}'
        return text



# (c) Roxy Corp. 2020-
# Roxy AI Inspect/Analyze-Server API
from __future__ import annotations
from .connection import Connection
from .com_definition import CommandCode
from .com_base import BaseCommand


class EchoCommand(BaseCommand):
    _CODE = CommandCode.ECHO

    def __init__(self, data: bytes, connection: Connection = None, logging: bool = True):
        """ ECHO コマンド制御

        Args:
            data:           送受信データ
            connection:     通信対象のTCP接続
            logging:        送受信時のログ出力フラグ
        """
        super().__init__(connection=connection, logging=logging)
        # 要求データの設定
        self.data = data

    def _decode_reply(self, reply: bytes):
        """ Inspect コマンドの応答内容確認

        Args:
            reply (bytes):      受信応答データ（ヘッダ以外）
        """
        # 応答データの妥当性チェック
        if self.data != reply:
            raise RuntimeError('mismatched echo reply data')

    def __str__(self) -> str:
        string = (
            f'{super().__str__()} '
            f'<-> Data: {self.data[:min(len(self.data), 128)]} '
            f'{len(self.data):,} '
        )
        return string



# (c) Roxy Corp. 2020-
# Roxy AI Coordinator
from __future__ import annotations
from typing import Callable, Any
from logging import getLogger
from time import perf_counter_ns

from .com_definition import CommandStatus

log = getLogger(__name__)


class ServerBaseHandler():
    code = None
    name = None

    # コマンド送受信ログ出力用
    last_command = None
    repeat_count = 0

    def __init__(
        self,
        reply_func: Callable[[CommandStatus, bytes, bytes], None],
        context: Any = None,
    ):
        """ 基底ハンドラクラス
        Args:
            reply_func:         応答処理関数
            context:            コマンド処理に必要な情報
        """
        # # ログ出力に手続き名を設定
        self._status = CommandStatus.STS_REQUEST
        self._reply_func = reply_func
        self.__start_time = None
        self.__end_time = None
        self._context = context

    def proceed(self):
        """ コマンドの処理呼び出し
        """
        try:
            # コマンド処理開始時刻の記録
            self.__start_time = perf_counter_ns()
            self.__end_time = None

            # 具象クラスの処理を呼び出し
            self.status = self.run()

            if self._status.is_ack:
                # 正常応答ならば具象クラスのデータ構築を呼び出し
                send_data = self.encode()
            else:
                send_data = b''

            # コマンド処理完了時間の記録
            self.__end_time = perf_counter_ns()

        except Exception:
            # コマンド処理中に例外が発生(TBD)
            self.status = CommandStatus.ERR_UNKNOWN_EXCEPTION
            send_data = b''
            log.exception(
                f"{self.__class__.__name__}: "
                f"Exception on command {self.code:02X}x procedure"
            )
        finally:
            # 応答処理
            self._reply_func(self._status, send_data)

        log.debug(f'command {self.__class__.__name__} replies {repr(self._status)}\n')

    def decode(self, payload: bytes):
        """ コマンド要求の受信データ解釈
        Note:
            要求パラメータのデコードを各具象クラスでオーバーライドする
        """
        # 処理するパラメータが無い場合は何も実施しない。
        pass

    def run(self):
        """ コマンドの処理実行
        Note:
            各具象クラスでオーバーライドする
        """
        raise NotImplementedError

    def encode(self):
        """ コマンド応答の送信データ構築
        Note:
            応答パラメータのデコードを各具象クラスでオーバーライドする
        """
        return b''

    @property
    def process_time(self) -> float:
        """ ハンドラの処理時間 [ms]
        """
        time_ms = None
        if self.__end_time and self.__start_time:
            time_ms = (self.__end_time - self.__start_time) / 1000000
        return time_ms

    def com_info(self):
        """ コマンドの情報文字列取得
        Note:
            各ハンドラの __str__ での利用を想定。
        """
        string = f"0x{self.code:02X}({self.code:d})"
        if self._status.is_reply:
            # コマンド応答時
            string += (
                f' [{str(self._status)}]'
            )
            if self._status.is_ack:
                # 正常応答時
                string += f'({self.process_time:,.2f} ms)'

        return string

    @property
    def status(self) -> CommandStatus:
        return CommandStatus(self._status)

    @status.setter
    def status(self, val: int):
        """ コマンドの送信／返信状態の設定
        """
        self._status = CommandStatus(val)



# (c) Roxy Corp. 2020-
# Roxy AI Inspect-Server API
from .connection import Connection
from .com_definition import (
    SIGN_CODE,
    HEADER_SIZE,

    CommandCode,
    CommandStatus,
    Judgment,
    JudgmentCount,
    JudgmentCounter,
    Probability,
    LabeledProbability,
    InspectResult,
)
from .inspect_image import (
    ImageFormat,
    ColorOrder,
    ImageBuffer,
    InspectImage,
    ImageConvertionFailed,
    ImageRect,
)
from .inspect_buffer import InspectBuffer
from .com_base import BaseCommand
from .com_echo import EchoCommand
from .com_initialize import InitializeCommand
from .com_terminate import TerminateCommand
from .com_inspect import InspectCommand
from .com_get_probabilities import GetProbabilitiesCommand
from .com_get_image import GetImageCommand

from .svr_base_handler import ServerBaseHandler
from .svr_command_handler import ServerCommandManager

from .label_info import (
    ProductLabelInfo, ModelLabelInfo, LabelManager
)

# Deprecated
# 過去APIバージョン互換のために定義
LabelInfo = ProductLabelInfo

__all__ = [
    Connection,
    BaseCommand,

    EchoCommand,
    InitializeCommand,
    TerminateCommand,
    InspectCommand,
    GetProbabilitiesCommand,
    GetImageCommand,

    CommandCode,
    CommandStatus,
    Judgment,
    JudgmentCount,
    JudgmentCounter,
    Probability,
    LabeledProbability,
    InspectResult,

    ImageFormat,
    ColorOrder,
    ImageBuffer,
    InspectImage,
    ImageConvertionFailed,
    ImageRect,

    InspectBuffer,

    ServerBaseHandler,
    ServerCommandManager,

    ModelLabelInfo,
    ProductLabelInfo,
    LabelManager,

    SIGN_CODE,
    HEADER_SIZE,

    # Deprecated
    LabelInfo
]
